## Filename    : chart_simul.py
## Author(s)   : Michel Le Borgne
## Created     : 04/2010
## Revision    :
## Source      :
##
## Copyright 2010 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall INRIA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux IRISA/IRSET
##
"""
Guarded transition systems simulator implementation
"""

from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor
from cadbiom.models.guard_transitions.simulator.translators.gt_visitors import\
                                      GenSimVisitor
from cadbiom.models.guard_transitions.simulator.simul_exceptions import\
                                      SimulException, SimulPropViolation
from cadbiom.models.guard_transitions.simulator.input_stream import CFileInput, ActInput
from simul_aux import ChartExpVisitor
from cadbiom.models.biosignal.translators.gt_visitors import compile_cond

# the simulator
class ChartSimulator(object):
    """
    wrap everything for simulation - this include the simulator model and simulator attribute to
    manage simulation.
    """
    def __init__(self):
        # the chart model which is the source of simulator model
        self.model = None

        # simulation model - components are described in chart_simul.py
        # first place to be activated at initialisation (graphical init)
        self.start = None
        self.__model_symb_tab = None
        self.__symb_tab = dict()   # symbol table
        # basic building blocks of a flat model (no macro state)
        self.__places = []
        self.__transitions = []    # other basic blocks
        self.__events = []
        self.__inputs = []
        self.input_names = []
        # list of __places which are activated at initialization
        self.init_places = []

        # properties
        self.invariant_prop = None
        self.reachable_prop = None

        # simulation helpers
        self.input_buffer = None # simulate __inputs described in file
        self.chrono = None       # auxiliary display"

        # simulator memory
        self.__activated_places = []
        self.__armed_transitions = []
        self.__fireabled_transitions = []
        self.__step = 0

    def build(self, chart_model, sim_option, reporter):
        """
        Given a chart_model, build a simulator
        @param chart_model: a ChartModel object
        @param sim_option: allows free clocks if True
        @param reporter: must be a CompilReporter
        """
        self.model = chart_model
        # symbol table built
        tvi = TableVisitor(reporter)
        chart_model.accept(tvi)
        if tvi.error:
            reporter.display("Error in symbol declarations")
            return
        # GenSimVisitor brings the simulator model with him and fill it
        # essentially flatten the initial model => no more macro
        self.__model_symb_tab = tvi.tab_symb
        gsv = GenSimVisitor(tvi.tab_symb, self, sim_option, reporter)
        chart_model.accept(gsv)

    def add_state_place(self, place, name):
        """
        state places are referenced in the simulator symbol table
        """
        self.__places.append(place)
        self.__symb_tab[name] = place

    def add_place(self, place):
        """
        add any place
        """
        self.__places.append(place)

    def add_transition(self, trans):
        """
        add a transition
        """
        self.__transitions.append(trans)

    def add_event(self, evt, name):
        """
        add an event
        """
        self.__events.append(evt)
        self.__symb_tab[name] = evt

    def add_input(self, inp):
        """
        add an input
        """
        self.__inputs.append(inp)
        self.input_names.append(inp.expression.name)

    def has_input(self):
        """
        test if the simulated model has inputs
        """
        return len(self.__inputs)>0

    def get_symb_tab(self):
        """
        the symbol table: name -> chart_simul_elem
        """
        return self.__symb_tab

    def get_step(self):
        """
        value of the simulator step counter
        """
        return self.__step


    def set_invariant_prop(self, prop):
        """
        set invariant property to be checked.
        A property violation exception is raised if violated
        @param prop: string - textual representation of property on places
        """
        reporter = SimReport()
        try:
            self.invariant_prop = compile_cond(prop, self.__model_symb_tab,
                                               reporter)
        except Exception , exc:
            raise SimulException("Error in property " + exc.__str__())

    def set_reachable_prop(self, prop):
        """
        set reachable property to be checked.
        @param prop: string - textual representation of property on places
        """
        reporter = SimReport()
        try:
            self.reachable_prop = compile_cond(prop, self.__model_symb_tab,
                                               reporter)
        except Exception , exc:
            raise SimulException("Error in property " + exc.__str__())


    def set_sce_file_input_stream(self, filename):
        """
        Find the file containing  description of a scenario
        set the input stream to read the __inputs only
        Throw SimulException
        """
        self.input_file = filename
        try:
            self.input_buffer = ActInput(filename)
            init_places = self.input_buffer.input_buffer[0].split()
            self.simul_init_places(init_places)
        except SimulException, sexc:
            self.input_buffer = None
            raise SimulException(sexc.message)

        except Exception, exc:
            self.input_buffer = None
            raise SimulException("Problem reading input file: " + exc.__str__())

    def set_act_file_input_stream(self, filename):
        """
        Find the file containing input description as an enumeration of activated
        __inputs  - set the input stream
        Throw SimulException
        """
        self.input_file = filename
        try:
            self.input_buffer = ActInput(filename)
        except SimulException, sexc:
            self.input_buffer = None
            raise SimulException(sexc.message)

        except Exception, exc:
            self.input_buffer = None
            raise SimulException("Problem reading input file:" + exc.__str__())

    def set_act_input_stream(self, act_input_list):
        """
        Set the input stream from a list of activated __inputs in format "% i1 i2 h2 h4 h8"
        """
        self.input_buffer = ActInput(inlist=act_input_list)


    def set_cfile_input_stream(self, filename):
        """
        Find the file containing input description using a language - set the input stream
        Throw SimulException
        """
        self.input_file = filename
        try:
            self.input_buffer = CFileInput(filename)
        except SimulException, sexc:
            self.input_buffer = None
            raise SimulException(sexc.message)

        except Exception, exc:
            self.input_buffer = None
            raise SimulException("Problem reading input file:" + exc.__str__())
        # check coherence with model
        inames = self.input_buffer.input_names
        nb_inames = self.input_buffer.nb_inputs
        if nb_inames == len(self.input_names):
            for name in inames:
                if not name in self.input_names:
                    self.input_buffer = None
                    raise SimulException("Name in file is not a model input")
        else:
            self.input_buffer = None
            mess = "Numbers of __inputs different in input file and model"
            raise SimulException(mess)

    # simulation methods
    def __set_armed_transitions(self):
        """
        At the beginning of the simulation we select __transitions with activated origin
        """
        self.__armed_transitions = []
        for place in self.__places:
            if place.activated:
                for trans in place.out_transitions:
                    self.__armed_transitions.append(trans)

    def __set_events(self):
        """
        Read the input buffer if any and set input __events
        """
        # set events to unactivated
        for evt in self.__events:
            evt.set_activated(False)
        # read input if any
        if self.input_buffer:
            nin = self.input_buffer.next_input()
            for evn in nin:
                try:
                    self.__symb_tab[evn].set_activated(True)
                except KeyError:
                    raise SimulException('Simul set_event: Unknown event')

    def __add_fireable_transitions(self):
        """
        A transition is fire-able if both its event and its condition are true
        """
        self.add_fir_trans = False
        expv = ChartExpVisitor(self.__symb_tab)
        for trans in self.__armed_transitions:
            if trans.event: # eval event
                ev_val = trans.event.expression.accept(expv)
            else:
                ev_val = True
            if ev_val: # event present
                if trans.condition: # eval condition
                    cond_val = trans.condition.expression.accept(expv)
                else:
                    cond_val = True
                if cond_val:
                    self.__fireabled_transitions.append(trans)
                    self.__armed_transitions.remove(trans)
                    self.add_fir_trans = True
                    # update __events
                    if trans.signal:
                        trans.signal.activated = True
                    # see if target and origin states emit an event
                    if trans.target.ev_up:
                        trans.target.ev_up.activated = True
                    if trans.origin.ev_down:
                        trans.origin.ev_down.activated = True

    def __set_fireable_transitions(self):
        """
        Fix point calculation of fire-able __transitions
        """
        self.__fireabled_transitions = []
        self.add_fir_trans = True
        while self.add_fir_trans:
            self.__add_fireable_transitions()

    def __fire_desactivate(self):
        """
        unactivation when transitions are fired
        """
        for trans in self.__fireabled_transitions:
            trans.desactivate()

    def __fire_activate(self):
        """
        activation when transitions are fired
        """
        for trans in self.__fireabled_transitions:
            trans.activate()


    def check_inv_conditions(self):
        """
        evaluate invariant property in the current state
        Raise a SimulPropViolation exception if invariant property is
        not satisfied.
        """
        if self.invariant_prop:
            expv = ChartExpVisitor(self.__symb_tab)
            val = self.invariant_prop.accept(expv)
            if not val:
                raise SimulPropViolation("Invariant violation",
                                         self.invariant_prop.__str__())

    def check_reachable_property(self):
        """
        evaluate final property in the current state
        @return: the result (bool) of the evaluation
        """
        if self.reachable_prop:
            expv = ChartExpVisitor(self.__symb_tab)
            val = self.reachable_prop.accept(expv)
            return val

    def simul_step(self):
        """
        Perform a simulation step
        """
        self.__set_armed_transitions()
        self.__set_events()
        self.__set_fireable_transitions()
        self.__fire_desactivate()
        self.__fire_activate()
        self.check_inv_conditions()
        self.model.notify()
        self.__step += 1

    def simul_step_trace(self, trace_file = None):
        """
        same as simul_step but register a trace
        """
        self.__set_armed_transitions()
        self.__set_events()
        self.__set_fireable_transitions()
        self.__fire_desactivate()
        self.__fire_activate()
        # __events are registered a step late
        if not trace_file:
            pln = []
            for place in self.__places:
                if place.activated and not place.is_input_place():
                    pln.append(place.name)
            for evt in self.__events:
                if evt.is_input_event() and evt.activated:
                    pln.append(evt.expression.__str__())
            print pln
        else:
            trace_file.write("\n%")
            for place in self.__places:
                if place.activated and not place.is_input_place():
                    trace_file.write(" "+place.name)
            for evt in self.__events:
                if evt.is_input_event() and evt.activated:
                    trace_file.write(evt.expression.__str__())
        self.check_inv_conditions()
        self.__step += 1

    def simul_init(self):
        """
        Simulator initialization
        """
        self.__step = 0
        self.model.clean()
        self.init_places = None
        for place in self.__places:
            place.force_desactivate()
        for str in self.__transitions:
            str.init_desactivate()
        for evt in self.__events:
            evt.set_activated(False)


    def simul_init_places(self, init_places):
        """
        @param init_places: list of place names to be activated
        """
        self.__step = 0
        self.model.clean()
        self.init_places = init_places
        for place in self.__places:
            if place.perm:
                place.activate()
            else:
                place.desactivate()
        for pl_name in init_places:
            self.__symb_tab[pl_name].activate()
        for str in self.__transitions:
            str.init_desactivate()
        self.check_inv_conditions()
        self.model.notify()

class SimReport(object):
    """
    Simple reporter used for simulation
    """

    def __init__(self):
        self.mess = ""

    def display(self, mess):
        """
        Interrupt all in case of problems
        """
        raise SimulException(mess)



# -*- coding: utf-8 -*-
## Filename    : TestCadlangCompiler.py
## Author(s)   : Michel Le Borgne
## Created     : 31/1/2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
##     http:
##     mailto:
##
## Contributor(s): Geoffroy Andrieux
##
from __future__ import print_function
import sys
import unittest
import pkg_resources
from antlr3 import ANTLRFileStream, CommonTokenStream

from cadbiom.models.guard_transitions.translators.cadlangLexer import cadlangLexer
from cadbiom.models.guard_transitions.translators.cadlangParser import cadlangParser
from cadbiom.models.guard_transitions.translators.chart_lang import LangVisitor
from cadbiom.models.guard_transitions.chart_model import ChartModel

class Reporter(object):
    def __init__(self):
        self.error = False
        self.mess = ""
        pass

    def display(self, hdr, mes=''):
        self.error = True
        self.mess = self.mess + "ERROR -> "+hdr+' '+mes
        print(self.mess)



class CadlangTest(unittest.TestCase):

    def testTransition(self):
        """
        model with one complete transition
        """
        err = Reporter()
        filename = pkg_resources.resource_filename(
            __name__, # package name
            "tests/test1.cal"
        )
        r = ANTLRFileStream(filename)
        lexer = cadlangLexer(r)
        lexer.set_error_reporter(err)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(err)
        model = ChartModel('test1')
        parser.cad_model(model)
        dc = LangVisitor(sys.stdout)
        parser.model.accept(dc)
        # True if there is an error
        res = err.error
        self.assert_(not res, 'Error in testTransition')

    def testComplexTransition(self):
        """
        model with one complete transition
        """
        err = Reporter()
        filename = pkg_resources.resource_filename(
            __name__, # package name
            "tests/testPID3.cal"
        )
        r = ANTLRFileStream(filename)
        lexer = cadlangLexer(r)
        lexer.set_error_reporter(err)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(err)
        model = ChartModel('test1')
        parser.cad_model(model)
        dc = LangVisitor(sys.stdout)
        parser.model.accept(dc)
        # True if there is an error
        res = err.error
        self.assert_(not res, 'Error in testComplexTransition')

    def testClock1(self):
        """
        model with transitions and complex clocks
        """
        err = Reporter()
        filename = pkg_resources.resource_filename(
            __name__, # package name
            "tests/testclock1.cal"
        )
        r = ANTLRFileStream(filename)
        lexer = cadlangLexer(r)
        lexer.set_error_reporter(err)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(err)
        model = ChartModel('testclock1')
        parser.cad_model(model)
        dc = LangVisitor(sys.stdout)
        parser.model.accept(dc)
        # True if there is an error
        res = err.error
        self.assert_(not res, 'Error in testClock1')

    def testConstraints1(self):
        """
        model with transitions, complex clocks and constraints
        """
        err = Reporter()
        filename = pkg_resources.resource_filename(
            __name__, # package name
            "tests/testconst1.cal"
        )
        r = ANTLRFileStream(filename)
        lexer = cadlangLexer(r)
        lexer.set_error_reporter(err)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(err)
        model = ChartModel('testconst1')
        parser.cad_model(model)
        dc = LangVisitor(sys.stdout)
        parser.model.accept(dc)
        # True if there is an error
        res = err.error
        self.assert_(not res, 'Error in testConstraints1')


if __name__ == "__main__":
    #import sys;sys.argv = ['', 'Test.testName']
    #unittest.main()

    suiteFew = unittest.TestSuite()
    #suiteFew.addTest(Testgt2Sig("testNoCondOnOutEvents"))
    #suiteFew.addTest(Testgt2Sig("testOneTransCondEvent"))
    suiteFew.addTest(CadlangTest("testConstraints1"))
    unittest.TextTestRunner(verbosity=2).run(suiteFew)



##
## Filename    : SysExpr.py
## Author(s)   : Michel Le Borgne
## Created     : 11-12/2008
## Revision    : 
## Source      : 
##
## Copyright 2008 - 2009 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA/IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE 
##     
##
## Contributor(s): Geoffroy Andrieux
##
"""
Standard representation of biosignal expressions as trees.
In addition to standard logical operator, we introduce the default
and when operators.
"""
# representation of signal expressions as trees

class SigExpression(object):
    """
    Generic abstract class
    """
    def __init__(self):
        pass
    def __str__(self):
        pass
    def get_signals(self):
        """
        Gives the set of idents used in the expression
        """
        return set([])
    
    def get_ultimate_signals(self, mcsystem):
        """
        Gives the set of signals used in the expression
        """
        return set([])
    
    def is_clock(self, symb_table, method=None):
        """
         Check if an expression always give a clock 
        """
        return False
    
    def is_ident(self):
        """
        As it says
        """
        return False
    
    def is_bot(self):
        """
        As it says
        """
        return False
    
    def is_const(self):
        """
        As it says
        """
        return False
    
    def is_const_false(self):
        """
        As it says
        """
        return False
    
    def accept(self, visitor):
        """
        Method for all kind of visitors - raise exception if not implemented
        """
        raise TypeError(self.__class__.__name__+": Not implemented")
    
    def test_equal(self, model):
        """
        test if the expression tree is equal to model tree (for TESTS)
        """
        pass
     
class SigIdentExpr(SigExpression):
    """
    Ident expressions are represented by their name. So we need a symbol
    table for type checking
    """
    def __init__(self, name):
        """
        @param name: the identifier 
        """
        self.name = name
        
    def __str__(self):
        return self.name
    
    def is_ident(self):
        return True
    
    def get_signals(self):
        return set([self.name])
    
    def get_ultimate_signals(self, mcsystem):
        try:
            scg = mcsystem.symbol_table[self.name]
        except:  # undefined component considered as ultimate
            return set([self.name])
        if scg.is_state():
            return set([self.name])
        elif scg.is_signal():
            sce = scg.expression
            if not sce: # undefined signal
                return set([self.name])
            else:
                return sce.get_ultimate_signals(mcsystem)
        else: # ???
            return set([self.name])
        
    def is_clock(self, symb_table, method=None):
        try:
            scg = symb_table[self.name]
            if not method:
                return scg.is_clock()
            else:
                return method(scg)
        except KeyError:
            return False

    def accept(self, visitor):
        return visitor.visit_sig_ident(self)
        
    def test_equal(self, model):
        if not isinstance(model, SigIdentExpr):
            return False
        else:
            return (self.name == model.name)
        
        
class SigConstExpr(SigExpression):
    """
    Constant signals: True, False
    """
    def __init__(self, val):
        """
        @param val: boll - the value 
        """
        self.value = val # boolean
        
    def __str__(self):
        return self.value.__str__()

    def get_signals(self):
        return set([])
    
    def is_clock(self, symb_table, method=None):
        """
        Tick and true are considered as similar
        """
        return self.value
    
    def is_const(self):
        return True
    
    def is_const_false(self):
        return not self.value
    
    def accept(self, visitor):
        return visitor.visit_sig_const(self)
        
    def test_equal(self, model):
        if not isinstance(model, SigConstExpr):
            return False
        else:
            return (self.value == model.value)
        
class SigBotExpr(SigExpression):
    """
    Usefull signal for initialisations - should not appear in normal signal expressions
    """
    def __init__(self):
        pass
        
    def __str__(self):
        return '%b'

    def get_signals(self):
        return set([])
    
    def is_bot(self):
        return True
    
    def accept(self, visitor):
        return visitor.visit_sig_bot(self)
        
    def test_equal(self, model):
        return isinstance(model, SigBotExpr)

class SigBinExpr(SigExpression):
    """
    Binary expression - abstract class
    """
    def __init__(self, exp1, exp2):
        """
        @param exp1,exp2: SigExpression - the two operands 
        """
        self.left_h = exp1
        self.right_h = exp2

    def __str__(self):
        pass
    
    def get_signals(self):
        sig1 = self.left_h.get_signals()
        sig2 = self.right_h.get_signals()
        sig1.update(sig2) #union of sets
        return sig1

    def get_ultimate_signals(self, mcsys):
        sig1 = self.left_h.get_ultimate_signals(mcsys)
        sig2 = self.right_h.get_ultimate_signals(mcsys)
        return sig1 | sig2 #union of sets
    
    
class SigDefaultExpr(SigBinExpr):
    """
    Implements the default operator
    """
    def __str__(self):
        if self.left_h.is_bot():
            return self.right_h.__str__()
        elif self.right_h.is_bot():
            return self.left_h.__str__()
        else:
            str = '('+self.left_h.__str__() + ' default '
            str = str  + self.right_h.__str__() + ')'
            return  str
    
    def is_clock(self, symb_table, method = None):
        cond = self.left_h.is_clock(symb_table, method)
        cond = cond and self.right_h.is_clock(symb_table, method)
        return  cond

    def accept(self, visitor):
        return visitor.visit_sig_default(self)

    def test_equal(self, model):
        if not isinstance(model, SigDefaultExpr):
            return False
        else:
            cond = self.left_h.test_equal(model.left_h)
            cond = cond and (self.right_h.test_equal(model.right_h))
            return cond
                    
class SigWhenExpr(SigBinExpr):
    """
    Implements when expression
    """
    def __str__(self):
        return '('+self.left_h.__str__()+' when '+self.right_h.__str__()+')'
    
    def is_clock(self, symb_table, method=None):
        return self.left_h.is_clock(symb_table, method)
    
    def accept(self, visitor):
        return visitor.visit_sig_when(self)
   
    def test_equal(self, model):
        if not isinstance(model, SigWhenExpr):
            return False
        else:
            cond = self.left_h.test_equal(model.left_h)
            cond = cond and (self.right_h.test_equal(model.right_h))
            return cond
        
class SigEqualExpr(SigBinExpr):
    """
    Implement the equality test: sig1 == sig2 when both present
    """
    def __str__(self):
        return '('+self.left_h.__str__()+' == '+self.right_h.__str__()+')'
    
    def is_clock(self, symb_table, method=None):
        cond = self.left_h.is_clock(symb_table, method)
        cond = cond and self.right_h.is_clock(symb_table, method)
        return  cond
    
    def accept(self, visitor):
        return visitor.visit_sig_equal(self)

    def test_equal(self, model):
        if not isinstance(model, SigEqualExpr):
            return False
        else:
            cond = self.left_h.test_equal(model.left_h)
            cond = cond and (self.right_h.test_equal(model.right_h))
            return cond
                    
class SigDiffExpr(SigBinExpr):
    """
    Implement the different test: sig1 != sig2 when both present
    """
    def __str__(self):    
        return '('+self.left_h.__str__()+' != '+self.right_h.__str__()+')'
    
    def is_clock(self, symb_table, method=None):
        return False
    
    def accept(self, visitor):
        return visitor.visit_sig_diff(self)

    def test_equal(self, model):
        if not isinstance(model, SigDiffExpr):
            return False
        else:
            cond = self.left_h.test_equal(model.left_h)
            cond = cond and (self.right_h.test_equal(model.right_h))
            return cond
        
class SigNotExpr(SigExpression):
    """
    Boolean not on a signal
    """
    def __init__(self, exp):
        self.operand = exp
        
    def __str__(self):
        return '( not '+self.operand.__str__()+')'

    def get_signals(self):
        return self.operand.get_signals()

    def get_ultimate_signals(self, mcsys):
        return self.operand.get_ultimate_signals(mcsys)

    def is_clock(self, symb_table, method=None):
        return False

    def accept(self, visitor):
        return visitor.visit_sig_not(self)

    def test_equal(self, model):
        if not isinstance(model, SigNotExpr):
            return False
        else:
            return (self.operand.test_equal(model.operand))
        
class SigEventExpr(SigExpression):
    """
    Event operand creates a clock
    """
    def __init__(self, exp):
        self.operand = exp
        
    def __str__(self):
        return 'event('+self.operand.__str__()+')'

    def get_signals(self):
        return self.operand.get_signals()
    
    def get_ultimate_signals(self, mcsys):
        return self.operand.get_ultimate_signals(mcsys)

    def is_clock(self, symb_table, method=None):
        return True

    def accept(self, visitor):
        return visitor.visit_sig_event(self)

    def test_equal(self, model):
        if not isinstance(model, SigEventExpr):
            return False
        else:
            return (self.operand.test_equal(model.operand))
                    
                    
class SigSyncBinExpr(SigBinExpr):
    """
    Generic class for boolean operation on signals
    A boolean operator emits a signal when both operand are present
    """
    def __init__(self, operator, exp1, exp2):
        self.left_h = exp1
        self.right_h = exp2
        self.operator = operator

    def __str__(self):
        lhs = self.left_h.__str__()
        rhs = self.right_h.__str__()
        return '('+ lhs +' '+self.operator+' '+rhs+')'
    
    def is_clock(self, symb_table, method=None):
        """
        Clocks are assimilated to boolean signals with true value
        And and or give true in this case, so its a clock!!
        """
        cond = self.left_h.is_clock(symb_table, method)
        cond = cond and self.right_h.is_clock(symb_table, method)
        return  cond

    def accept(self, visitor):
        return visitor.visit_sig_sync(self)

    def test_equal(self, model):
        if not isinstance(model, SigSyncBinExpr):
            return False
        else:
            cond = self.left_h.test_equal(model.left_h)
            cond = cond and (self.right_h.test_equal(model.right_h))
            cond = cond and self.operator == model.operator
            return cond

#class SigPolyBinExpr(SigBinExpr):
#    def __init__(self, op, exp1, exp2):
#        self.left_h = exp1
#        self.right_h = exp2
#        self.operator = op
#        
#    def __str__(self):
#        lhs = self.left_h.__str__()
#        rhs = self.right_h.__str__()
#        return '('+ lhs +' '+self.operator+' '+rhs+')'
#    
#    def is_clock(self, ts, method=None):
#        """
#        Clocks are assimilated to boolean signals with true value
#        In general a polynomial expression doesn't deliver a clock
#        """
#        return False
#    
#    def accept(self, visitor):
#        return visitor.visitSigPolyBinExpr(self)
#
#    def test_equal(self, model):
#        if not isinstance(model, SigPolyBinExpr):
#            return False
#        else:
#            cond = self.operator==model.operator
#            cond = cond and (self.left_h.test_equal(model.left_h)
#            cond = cond and (self.right_h.test_equal()))
#            return cond
#                            
#class SigPolyPowExpr(SigExpression):
#    def __init__(self, exp, pow):
#        self.operand = exp
#        self.power = pow%3
#        
#    def __str__(self):
#        op=self.operand.__str__()
#        return '('+ op +'^'+"%d"%(self.power)+')'
#
#    def get_signals(self):
#        return self.operand.get_signals()
#
#    def get_ultimate_signals(self, mcsys):
#        return self.operand.get_ultimate_signals(mcsys)
#
#    def is_clock(self, ts, method=None):
#        if self.power == 2:
#            return True
#        else:
#            return False
#        
#    def test_equal(self, model):
#        if not isinstance(model, SigPolyPowExpr):
#            return False
#        else:
#            cond = self.power==model.power
#            cond = cond and (self.operand.test_equal(model.operand))
#            return  cond
#                    
#    def accept(self, visitor):
#        return visitor.visitSigPowExpr(self)

class SigConstraintExpr(SigExpression):
    """
    Implements constraints on events
    """
    SYNCHRO = 'synchro'
    EXCLU = 'exclus'
    INCL = 'included'
    
    def __init__(self, name, explist=[]):
        self.constraint_name = name
        self.arg = explist
        
    def add_arg(self, arg):
        """
        Constraints are n-ary operators - this method add an operand
        """
        self.arg.append(arg)
        
    def __str__(self):
        str_out = self.constraint_name + '('
        if len(self.arg)>0:
            str_out = str_out + self.arg[0].__str__()
        for arg in self.arg[1:]:
            str_out = str_out +', '+arg.__str__()
        str_out = str_out+')'
        return str_out
    
    def accept(self, visitor):
        return visitor.visit_sig_constraint(self)






## Filename    : CLDynSys.py
## Author(s)   : Michel Le Borgne
## Created     : 05/2011
## Revision    :
## Source      :
##
## Copyright 2011 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and  has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux
##
"""
Classes for building clause constraint models
"""
from collections import defaultdict


class Clause():
    """Objects representing logical clauses

    A clause is represented as a list of literals.

    Attributes:

        :param literals: Variable name
        :type literals: <list <Literal>>
    """
    def __init__(self, list_lit=[]):
        self.literals = list_lit

    def add_lit(self, lit):
        """Add a literal to the clause
        @param lit: <Literal>
        """
        self.literals.append(lit)

    def __eq__(self, other):
        """Test if two clauses have the same literals
        Used by clause_list_equal() (never used)
        """
        return frozenset(self.literals) == frozenset(other.literals)

    def __ne__(self, other):
        """Test clause inequality"""
        return not self.__eq__(other)

    def __lt__(self, other):
        """Test if the current clause is shorter than another"""
        return len(self.literals) < len(other.literals)

    def __gt__(self, other):
        """Test if the current clause is longer than another"""
        return len(self.literals) > len(other.literals)

    def __str__(self):
        return "$" + ", ".join(str(lit) for lit in self.literals) + "$"

    def __repr__(self):
        return str(self)


class Literal():
    """Object representing literals.

    A literal is a pair (string, boolean).

    Attributes:

        :param name: Variable name
        :param sign: Boolean the sign of the literal.
            Positive or negative: True or False
        :type name: <str>
        :type sign: <boolean>
    """
    def __init__(self, name, sign):
        self.name = name
        self.sign = sign

    def lit_not(self):
        """Return the negation of the literal (opposite sign)"""
        return Literal(self.name, not self.sign)

    def __eq__(self, other):
        """Test literal equality (never used except in tests)"""
        return (other.name == self.name) and (other.sign == self.sign)

    def __ne__(self, other):
        """Test literal inequality"""
        return not self.__eq__(other)

    def __hash__(self):
        """Used for set operations"""
        return hash(self.name) ^ hash(self.sign)

    def opposite(self, other):
        """true if one literal is the negation of the other"""
        return (other.name == self.name) and (other.sign == (not self.sign))

    def code_name(self):
        """Return a string representing the literal"""
        return self.name if self.sign else '_' + self.name

    def __str__(self):
        return self.name if self.sign else 'not ' + self.name

    def __repr__(self):
        return str(self)


class CLDynSys(object):
    """Class to describe a dynamic system in clause form.

    :param symb_tab: Symbol table of charter model produced by Table_visitor
        Dict of entities names as keys, and states as values.
        States are described like: (type, deepness); where deepness if the
        imbrication level of macro states (Cf Table_visitor).

        All keys = no_frontiers + frontiers + model name
        (probably a bug for this last one)

        /!\ Should not be used after adding the clauses by GT2Clauses (?)

        :Examples:
            {'PGK_cGMP_active': ('state', 0), ...}
            {'input_place': ('input', 0), ...}

        :Completed by MCLTranslator with add_free_clock():
            {"_h_0000": ('clock', -1)}

    :param report: Reporter for error reporting.
    :param base_var_set: Set of ALL variables of the dynamic system
        (including inputs, entities, clocks/events, auxiliary variables)
    :param list_clauses: Clauses
        Added from MCLTranslator.
    :param aux_list_clauses: Auxiliary clauses.
        Added from MCLTranslator.
    :param frontiers: All frontiers of the model.
        Used by CLUnfolder.
        Added from MCLTranslator.
        Cf GT2Clauses.visit_csimple_node()
    :param no_frontiers: Places that are not frontiers.
        Added from MCLTranslator.
        Cf GT2Clauses.visit_csimple_node()
    :param free_clocks: Free clocks names inputs.
        Added from MCLTranslator.
    :param place_clocks: Dictionary of clocks as keys and places as values.
        Association between a clock and an inductive place:
            clock h -> [place P, ...] where P is a place preceding free clock h

        :Example:
            {"_h_2772": ["BTG2_gene", ...]}

    :param inputs: Other inputs
    :param lit_compt: Incrementable id for auxiliary variables naming.

    :type symb_tab: <dict <str>: <tuple <str>, <int>>>
    :type report: <Reporter>
    :type base_var_set: <set <str>>
    :type list_clauses: <list <Clause>>
    :type aux_list_clauses: <list <Clause>>
    :type frontiers: <list <str>>
    :type no_frontiers: <list <str>>
    :type free_clocks: <list <str>>
    :type place_clocks: <defaultdict <list>>
    :type inputs: <list <str>>
    :type lit_compt: <int>

    The frontiers, free_clocks, place_clocks lists are used for extraction of
    informations from solutions.
    The place_clocks list is used for generating the structural constraints
    OR(place P and clock h) implying that at least one clock transition must
    be activated at each step.

    Additional information about attributes:

        Model Entities: Their names are preserved from the model.
            Added to base_var_set
        Auxiliary variables: They are named by adding a value incremented at their end.
            Added to base_var_set
            Ex: "_lit00000"
        Free clock variables: Their names are preserved from the model.
            Added to base_var_set
            Added to free_clocks
            Added to the dict symb_tab:
                Ex: {"_h_0000": ('clock', -1)}
        Place clocks:
            Added to place_clocks:
                Ex: {"_h_2772": ["BTG2_gene", ...]}
        Input variables: Their names are "supposed to be" preserved from the model.
            Added to base_var_set
            Added to inputs (May be added multiple times ?)
        Clauses:
            Added to list_clauses
        Auxiliary clauses:
            Added to aux_list_clauses
    """
    def __init__(self, symb_tab, report):
        # symbol table of charter model produced by Table_visitor
        # completed by add_free_clock
        self.symb_tab = symb_tab
        self.report = report           # reporter for errors
        # Set of ALL variables of the dynamic system (including auxiliary ones).
        self.base_var_set = set()
        self.list_clauses = []         # clause form of the dynamic
        # structural constraints valid if there is no timing constraints
        self.aux_list_clauses = []
        self.frontiers = []            # frontiers places of the model
        self.no_frontiers = []         # Places that are not frontiers

        self.free_clocks = []          # free clocks inputs
        # dictionary of clocks as keys and places as values
        # association between a clock and an inductive place
        # clock h -> [place P, ...] where P is a place preceding free clock h
        self.place_clocks = defaultdict(list)
        self.inputs = []               # other inputs

        self.lit_compt = 0             # for auxiliary variables generation

    def add_var(self, name):
        """add a logical variable to the dynamic system
        @param name: the name of the variable (string)
            All entities of the model are added here
        """
        if name in self.base_var_set:
            return
        self.base_var_set.add(name)

    def add_aux_var(self):
        """create an auxiliary variable for compiler purpose

        Auxiliary variables are named by adding a value incremented at their end
            ex: "_lit00000"
        """
        name = '_lit' + str(self.lit_compt)
        self.lit_compt += 1
        self.base_var_set.add(name)
        return name

    def add_free_clock(self, hname):
        """add a free clock variable
        @param hname: the name of the clock variable (string)
            Free clocks are associated to the value ('clock', -1) in self.symb_tab
            ex: {"_h_0000": ('clock', -1)}

        """
        if hname in self.base_var_set:
            return
        self.base_var_set.add(hname)
        self.free_clocks.append(hname)
        self.symb_tab[hname] = ('clock', -1)

    def add_place_clock(self, pname, hname):
        """
        add an association hname --> pname between a clock and an inductive place
        """
        self.place_clocks[hname].append(pname)

    def add_input(self, name):
        """add a variable representing an input
        @param name: name of the input
        """
        if not name in self.base_var_set:
            self.base_var_set.add(name)
        self.inputs.append(name) # May be added multiple times ?

    def add_clause(self, cla):
        """
        add a clause constraint
        """
        #TODO: do not add satisfied clauses (n or not n)
        self.list_clauses.append(cla)

    def add_aux_clause(self, cla):
        """
        add an auxiliary clause constraint
        """
        self.aux_list_clauses.append(cla)

    def get_var_number(self):
        """Get the number of ALL variables of the dynamic system
        (including inputs, entities, clocks/events, auxiliary variables)
        """
        return len(self.base_var_set)




# -*- coding: utf-8 -*-
## Filename    : MCLSolutions.py
## Author(s)   : Michel Le Borgne
## Created     : 03/9/2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux - IRISA/IRSET
##
"""Classes used to store solver answers.

RawSolution:
    Contain a solution got from SAT solver with all variable parameters
    from the unfolder.

DimacsFrontierSol:
    A numerical representation of *frontier values and timings* from a raw solution.

FrontierSolution:
    Provides is a wrapper for a symbolic
    representation (human readable) of activated frontiers defined in
    RawSolution and DimacsFrontierSol objects.
"""
# Standard imports
from __future__ import print_function
import itertools as it
#from functools import cached_property
from cached_property import cached_property

from cadbiom import commons as cm

# C++ API
from _cadbiom import get_unshift_code, unflatten

LOGGER = cm.logger()


class MCLException(Exception):
    """Exception for MCLAnalyser"""
    def __init__(self, mess):
        self.message = mess

    def __str__(self):
        return self.message

################################################################################

class RawSolution(object):
    """RawSolution objects contain a solution got from SAT solver with all
    variable parameters from the unfolder.

    The class RawSolution provides a representation of solutions which are
    raw results from the SAT solver with all variable parameters from the
    unfolder, and consequently not easily readable.

    In addition to the solution in DIMACS form (including values for auxiliary
    variables), a RawSolution object registers data which permit information
    extraction from these raw solutions. The most important methods/attributes are:

        - frontier_pos_and_neg_values:
            Extracts the frontier values from the solution.
            These values are in DIMACS code.

        - extract_activated_frontier_values():
            Extracts frontier variables which are active in solution.

        - extract_act_input_clock_seq():
            Extract the sequence of activated inputs and events in the solution.

    ..warning:: IMPORTANT: the solution MUST correspond to the properties registered
        for its computation. Any change in these properties will give incorrect
        results (current_step, etc.).

    Attributes:
        :param shift_direction: Unfolding direction
        :param shift_step: The shift step ss (if n is X_0 code, n+ss is X_1 code) (?)
            Dependant on the run.
        :param solution: DIMACS encoding of **a solution** from the solver
            A solution is list of state vectors (tuple of literals).
            Each literal is a variable.
            ..note:: The solver returns list of solutions.
        :param unfolder: Get the current unfolder for the solution
            (unfolder is used for invariant constants)
        :param current_step: Get current step of the unfolder
            (Thus the current number of steps related to the current solution and
            by extension the maximum of steps until now)
        :type shift_direction: "FORWARD" or "BACKWARD"
        :type shift_step: <int>
        :type solution: <tuple <int>>
        :type unfolder: <CLUnfolder>
        :type current_step: <int>
    """
    def __init__(self, solution, unfolder):
        """
        :param solution: DIMACS encoding of **a solution** from the solver
            (tuple of literals)
        :param unfolder: Current unfolder for the solution
            (unfolder is used for invariant constants)
        :type solution: <tuple <int>>
        :type unfolder: <CLUnfolder>
        """
        self.solution = solution
        self.unfolder = unfolder
        self.shift_step = unfolder.get_shift_step()
        self.current_step = unfolder.get_current_step()
        self.shift_direction = unfolder.get_shift_direction()

    def get_unshift_code(self, var_num):
        """Get the real value of the given variable in the system (remove the shift)

        ..warning:: DEPRECATED Used internally by self.unflatten()

        @param var_num: DIMACS literal coding of a shifted variable x_i
        @return: DIMACS literal coding of  x_0 with same value
        :return: <int>
        """
        # Old API
        # var_code = (abs(var_num) - 1) % self.shift_step + 1
        #
        # if var_code <= self.shift_step:
        #     return var_code * (-1 if var_num < 0 else 1)
        #
        # raise MCLException("Not a DIMACS code of an initial variable")

        # New API via C++ module
        get_unshift_code(var_num, self.shift_step)

    def get_var_name(self, var_num):
        """For translation to symbolic representation
        Return the name of a variable with its value.
        """
        return self.unfolder.get_var_name(var_num)

    def unflatten(self):
        """Extract trajectories from the current solution

        Transform a flat DIMACS representation of a trajectory into a
        <list <list <int>>>.
        Each sublist represents a state vector (the state of all variables of
        the system for 1 step in the trajectory).

        ..note:: This assertion seems to be not true:
            Auxiliary variables introduced by properties compiling are ignored.
            => there are lots of variables on the format "_name"...

        ..note:: Debug with:

            >>> raw_sol = RawSolution()
            >>> print(raw_sol)

        ..note:: The list returns all variables of the system including
            inputs, clocks, auxiliary variables.

        :return: A list of state vectors (<list <int>>) in DIMACS format
        :rtype: <list <list <int>>>
        """

        # Old API
        # lv_traj = []
        # step = 0
        # dec = 0
        # shift_step_init = self.unfolder.shift_step_init
        # # n steps means h0, ..., h(n-1) in constraints for clocks and inputs
        # while dec + self.shift_step < len(self.solution):
        #     #assert [self.get_unshift_code(self.solution[i+dec]) for i in range(shift_step_init)] == [get_unshift_code(self.solution[i+dec], self.shift_step) for i in range(shift_step_init)]
        #     lv_traj.append(
        #         [get_unshift_code(self.solution[i+dec], self.shift_step)
        #          for i in range(shift_step_init)]
        #     )
        #     step += 1
        #     dec = step * self.shift_step # first index of time step
        # return lv_traj # list of lists

        # New API via C++ module
        return unflatten(
            self.solution,
            self.unfolder.shift_step_init,
            self.shift_step
        )

    def extract_activated_frontier_values(self):
        """Extracts frontier variables which are active in solution.

        :return: Set of activated frontier values.
        :rtype: <set>
        """
        if self.shift_direction == 'FORWARD':
            # frontier_values: Set of frontier positive values only.
            return frozenset(self.unfolder.frontier_values) & self.frontier_pos_and_neg_values
        else:
            NotImplementedError("BACKWARD shift direction is not yet implemented")

    @cached_property
    def frontier_pos_and_neg_values(self):
        """Extracts the frontier values from the solution.
        These values are in DIMACS code.

        .. warning:: This function returns values of active AND inactive frontiers.

        .. warning:: Assert: solution is sorted by variable code for BACKWARD

        .. note:: This function makes set operations between all values of
            frontiers and the current solution.
            This last one is a list of hundreds of thousands items
            (ex: ~500000 for a solution with 10 steps).
            Thus, the operation is costly and we use a decorator @cached_property
            to put the result in cache.

        .. note:: Used by:
            - self.extract_activated_frontier_values()
            - DimacsFrontierSol.__init__()

        .. note::
            frontiers (fixed size): 3824
            solution (increasing size): 171231, 228314, 285405, 285453, 400091 etc.

        @param sol: a dimacs solution
        @return: a dimacs code list of the activation state on the frontier for sol
        """
        # Old API
        # jmax = len(self.unfolder.frontier_values)
        # fsol = []
        # j = 0
        # dec = self.shift_step * self.current_step
        # sol = self.solution
        # for i in range(len(sol)):
        #     # the variable is frontier var:
        #     if abs(sol[i]) == self.unfolder.frontier_values[j]:
        #         if self.shift_direction == "FORWARD":
        #             fsol.append(sol[i])
        #         else:
        #             fsol.append(sol[i+dec])     # look for initial values
        #         j += 1
        #         if j == jmax:
        #             return fsol
        # return fsol # never reach this point

        # Old API v2
        # frontiers = self.unfolder.frontier_values
        # jmax = len(frontiers)
        # fsol = list()
        # j = 0
        # dec = self.shift_step * self.current_step
        # for i, sol in enumerate(self.solution):
        #    # the variable is frontier var:
        #    if abs(sol) == frontiers[j]:
        #        if self.shift_direction == "FORWARD":
        #            fsol.append(sol)
        #        else:
        #            fsol.append(self.solution[i+dec])# look for initial values
        #        j += 1
        #        if j == jmax:
        #            return fsol
        #return fsol # never reach this point

        if self.shift_direction == 'FORWARD':
            # frontiers_pos_and_neg:
            # All frontiers and their opposite version Union
            # Set of frontier positive and negative values.
            return self.unfolder.frontiers_pos_and_neg & frozenset(self.solution)
        else:
            # NEVER USED (BACKWARD is partially implemented)
            frontiers = self.unfolder.frontier_values
            jmax = len(frontiers)
            fsol = list()
            j = 0
            dec = self.shift_step * self.current_step
            frontier = frontiers[0]
            for i, var in enumerate(self.solution):
                # the variable is frontier var:
                if abs(var) == frontier:
                    fsol.append(self.solution[i+dec])# look for initial values
                    j += 1
                    if j == jmax:
                        # all frontiers are found
                        return fsol
                    frontier = frontiers[j]
        return fsol # never reach this point

    # Useless proposition
    # ps: jmax allows to stop the search when the iteration is made on solution
    # which is increasing over time.
    # without jmax, this proposition is not efficient
    #        frontiers = frozenset(self.unfolder.frontier_values)
    #        dec = self.shift_step * self.current_step
    #        if self.shift_direction == 'FORWARD':
    #            fsol = [sol for sol in self.solution if abs(sol) in frontiers]
    #        else:
    #            fsol = [self.solution[i+dec] for i, sol in enumerate(self.solution) if abs(sol) in frontiers]
    #
    #        assert len(fsol) == len(frontiers)
    #        return fsol

    def extract_act_inputs_clocks(self, state_vector):
        """Extract active *inputs* and *free_clocks* from a state vector.

        Return the intersection of all inputs and free clocks of the system
        with the given state vector.

        PS: unfolder.inputs_and_free_clocks contains only positive values
            so the intersection with state_vector allows to get only activated
            variables.

        Used only by extract_act_input_clock_seq().

        :param state_vector: A state vector. A list of literals.
        :return: Tuple of activated inputs and free_clocks in the given state vector.
        :type state_vector: <list <int>>
        :rtype: <tuple>
        """
        # Old API
        # return [s_varcode for s_varcode in state_vector
        #        if s_varcode > 0
        #        and s_varcode in self.unfolder.inputs_and_free_clocks]
        return tuple(self.unfolder.inputs_and_free_clocks & frozenset(state_vector))


    def extract_act_input_clock_seq(self):
        """Extract the sequence of activated *inputs* and *events/free clocks*
        in the current solution.

        .. note::
            - Each tuple is the state of literals in a step of the solution.
            - Each literal in the tuple is an activated variable in a step.
            - Variables are inputs or events (free clocks) ONLY!

        .. note:: A solution is list of state vectors (list of literals).

        .. note:: This function is called by:
            - FrontierSolution.from_raw()
            - DimacsFrontierSol.ic_sequence

        :return: list of input vector
            (may be empty if data-flow model is without input)
        :rtype: <tuple <tuple <int>>>
        """
        # unflatten() returns a list of state vectors in DIMACS format
        # State vectors are (<list <int>>)
        return tuple() \
            if not self.unfolder.inputs_and_free_clocks \
            else tuple(
                self.extract_act_inputs_clocks(activated_inputs_and_clocks)
                for activated_inputs_and_clocks in self.unflatten() # list of lists
            )

    def __len__(self):
        """Length of solution (number of tuples of literals)
        @return: int
        """
        return len(self.solution)

    def __str__(self):
        """For debug: display all variables
        (including inactivated and auxiliary ones)
        """
        unfl = self.unflatten()
        stn = []
        for vect in unfl:
            vect_n = []
            for dcod in vect:
                if dcod > 0:
                    vect_n.append(self.get_var_name(dcod))
                else:
                    vect_n.append("-"+self.get_var_name(dcod))
            stn.append(vect_n)
        return str(stn)

    def display_activ_sol(self):
        """For debug: display only activated variables
        (auxiliary variables are pruned)
        """
        unfl = self.unflatten()
        stn = []
        for vect in unfl:
            vect_n = []
            for dcod in vect:
                if dcod > 0:
                    v_name = self.get_var_name(dcod)
                    if not v_name[0] == "_":
                        # Not auxiliary variable
                        vect_n.append(v_name)
            vect_n.sort()
            stn.append(vect_n)
        print(stn)


class FrontierSolution(object):
    """Class for symbolic (human readable) frontiers and timings representation

    The class FrontierSolution provides is a wrapper for a symbolic
    representation (human readable) of activated frontiers defined in
    RawSolution and DimacsFrontierSol objects.

    Attributes:
        - The main attributes are activated_frontier which is a set of
        (string) names of frontier places which are activated in the solution.

        - The second important attribute is ic_sequence, a list of strings
        which describes the successive activated inputs and free events in
        the solution.

        - current_step is the number of steps used during the unfolding that made
        the solution.

    If events h2, h5 and input in3 are activated at step k in the solution,
    then the kth element of the list is “% h2 h5 in3”.
    This format is understood by the Cadbiom simulator.

    :type activated_frontier: <frozenset <str>>
    :type ic_sequence: <list <str>>
    :type current_step: <int>

    ..TODO::
        - Faire une super classe dont héritent RawSolution et DimacsFrontierSol
        pour éviter la duplication de code et le stockage d'attributs qu'on
        connait déjà dans les types de plus bas niveaux...
        - ... Ou détecter le type d'objet dans le constructeur => ducktyping
        - renommer l'attr activated_frontier en activated_frontiers
    """

    def __init__(self, activated_frontiers, ic_sequence, current_step):
        """Build a symbolic representation of frontier values and timing from
        a raw solution

        :param activated_frontiers: A list of (string) names of frontier places
            which are activated in the solution.
        :param ic_sequence: A list of strings which describes the successive
            activated inputs and free events in the solution.
            (format: % h1 h2)
        :param current_step: Number of steps in the sequence.
            ..note:: This attribute should be deducted with the length of ic_sequence.
        :type activated_frontiers: <frozenset <str>>
        :type ic_sequence: <list <str>>
        :type current_step: <int>
        """
        self.activated_frontier = frozenset(activated_frontiers)
        self.ic_sequence = ic_sequence
        self.current_step = current_step

    @classmethod
    def build_input_clock_sequence(cls, get_var_name, ic_sequence):
        """Get strings representing timings of each step in the solution.

        :param get_var_name: Method to call for a conversion of values to names.
            (binding of the method get_var_name() of an unfolder).
        :param ic_sequence:
            - Each tuple is the state of literals in a step of the solution.
            - Each literal in the tuple is an activated variable in a step.
            - Variables are inputs or events (free clocks) ONLY!
        :type get_var_name: <function>
        :type ic_sequence: <tuple <tuple <int>>
        :return: List of strings representing timings of each step in ic_sequence.
        :rtype: <list <str>>
        """
        # Strip used to remove trailing space on empty steps : "% "
        return [
            str("% " + " ".join(get_var_name(value)
            for value in step_literals)).rstrip(" ")
            for step_literals in ic_sequence
        ]

    @classmethod
    def from_raw(cls, raw_sol):
        """Build FrontierSolution object, a symbolic representation of frontier
        values and timings from a RawSolution solution.

        - Extract activated frontiers values (literals)
        - Convert activated frontiers values to places names (strings)
        - Extract inputs and clocks sequence (steps of the whole solution)

        :type raw_sol: <RawSolution>
        """
        # Get names of activated frontier places
        # RawSolution has not activated_frontiers attr we must compute them
        activated_frontier_values = raw_sol.extract_activated_frontier_values()

        activated_frontiers = \
            {raw_sol.get_var_name(var_num) for var_num in activated_frontier_values}

        # Get raw activated inputs and free clocks (values)
        # and format them so that they are understandable by humans
        ic_sequence = FrontierSolution.build_input_clock_sequence(
            raw_sol.get_var_name,
            raw_sol.extract_act_input_clock_seq()
        )
        # LOGGER.debug("FrontierSolution:from_raw:: ic_sequence %s", ic_sequence)

        return cls(activated_frontiers, ic_sequence, raw_sol.current_step)

    @classmethod
    def from_dimacs_front_sol(cls, dimacs_front_sol):
        """Build FrontierSolution object, a symbolic representation of frontier
        values and timings from a DimacsFrontierSolution solution.

        - Convert activated frontiers values to places names (strings)
        - Extract inputs and clocks sequence (steps of the whole solution)

        :type dimacs_front_sol: <DimacsFrontierSol>
        """
        # Get names of activated frontier places
        # DimacsFrontierSol already has activated_frontiers attr
        activated_frontiers = \
            {dimacs_front_sol.get_var_name(var_num)
             for var_num in dimacs_front_sol.activated_frontier_values}

        # Get raw activated inputs and free clocks (values)
        # and format them so that they are understandable by humans
        ic_sequence = FrontierSolution.build_input_clock_sequence(
            dimacs_front_sol.get_var_name,
            dimacs_front_sol.ic_sequence
        )
        # LOGGER.debug("FrontierSolution:from_dimacs_front_sol:: ic_sequence %s", ic_sequence)

        return cls(activated_frontiers, ic_sequence, dimacs_front_sol.current_step)

    @classmethod
    def from_file(cls, file_name):
        """Build a symbolic representation of frontier values and timing from
        a DimacsFrontierSolution solution

        ..warning:: This method can only be used if the file contains 1 solution...

        Note: use cadbiom_cmd/tools/solutions.py functions to reload frontiers
        """
        with open(file_name, "r") as f_d:
            # Get first line: frontiers
            act_front = f_d.readline().split()
            # Get steps
            ic_seq = []
            step = f_d.readline()
            current_step = 0
            while step:
                current_step += 1
                ic_seq.append(step[:-1])
                step = f_d.readline()

        LOGGER.debug("FrontierSolution:from_file:: ic_seq: %s; maxsteps: %s",
                     ic_seq, current_step)
        return cls(act_front, ic_seq, current_step)

    def save(self, outfile, only_macs=False):
        """Save a symbolic solution in a file
        The format is readable by the simulator and all Cadbiom tools.

        .. note:: Frontiers are sorted in (lower case) alphabetical order.

        :param outfile: Writable file handler
        :param only_macs: Write only frontiers, and skip timings.
        :type outfile: <open file>
        :type only_macs: <boolean>
        """
        outfile.write(' '.join(sorted(self.activated_frontier, key=lambda s: s.lower())) + '\n')
        if not only_macs:
            outfile.write('\n'.join(self.ic_sequence) + '\n')
        outfile.flush()

    def __eq__(self, other):
        """Check if two FrontierSolution have same activated frontiers
        Used with __hash__

        ..warning:: We do not test other attributes than activated_frontier.
            ic_sequence might be necessary.

        :type other: <DimacsFrontierSol>
        :rtype: <boolean>
        """
        return self.activated_frontier == other.activated_frontier

    def __ne__(self, other):
        """Python 2 requires this in conjunction to __eq__..."""
        return self.__eq__(other)

    def __hash__(self):
        """Used to test unicity of FrontierSolution in sets;
        Used with __eq__

        ..warning:: We do not test other attributes than activated_frontier.
            ic_sequence might be necessary.
        """
        return hash(self.activated_frontier)

    def __str__(self):
        """String representation of a FrontierSolution"""
        stro = "<FrontierSolution - Activated frontiers: "
        stro += str(self.activated_frontier)
        stro += "\nTimings:\n   "
        stro += "\n   ".join(self.ic_sequence) + ">"
        return stro


class DimacsFrontierSol(object):
    """Class for solution frontier and timings representation

    DimacsFrontierSol is a numerical representation of *frontier values and timings*
    from a raw solution.

    Attributes:

        :param frontier_values: Frontier values from the solution.
            These values are in DIMACS code.
            This attribute is **immutable**.
            Based on the initial RawSolution.
        :param ic_sequence: Sequence of activated inputs and events in the solution.
            Only used in FrontierSolution.from_dimacs_front_sol()
            (at the end of the search process for the cast into FrontierSolution)
            and for tests.
            Based on the initial RawSolution.
        :param current_step: The number of steps allowed in the unfolding;
            the current_step on which the properties must be satisfied (never used)
            Based on the initial RawSolution.
        :param nb_activated_frontiers: Number of activated places in the current
            Dimacs solution.
            This attribute is precomputed because it is used intensively and
            multiple times for each object in:
                MCLAnalyser.less_activated_solution
                MCLAnalyser.__prune_frontier_solution
                MCLAnalyser.__mac_exhaustive_search
        :param activated_frontier_values: Values of activated frontier places
            in the current Dimacs solution.
            Convenient attribute used:
                - by FrontierSolution.from_dimacs_front_sol()
                - to accelerate the call of self.nb_activated_frontiers
                - as a criterion of uniqueness, to do set operations
        :type frontier_values: <frozenset <int>>
        :type ic_sequence: <list <list <int>>>
        :type current_step: <int>
        :type nb_activated_frontiers: <int>
        :type activated_frontier_values: <frozenset <int>>
    """

    def __init__(self, raw_sol):
        """Build a DimacsFrontierSol, a numeric representation of frontier values
        and timings from a raw solution

        :param raw_sol: raw solution
        :type raw_sol: <RawSolution>
        """
        # Frontier places
        # /!\ Most of the time spent in the constructor is due to this line
        # Already a frozenset but this attribute must be immutable for __hash__,
        # cast is a security.
        self.frontier_values = frozenset(raw_sol.frontier_pos_and_neg_values)

        # Activated input/events (clocks)
        # Callable of raw_sol method (avoid to store raw_sol in this object)
        self._ic_sequence_func = raw_sol.extract_act_input_clock_seq
        self.current_step = raw_sol.current_step

        # CLUnfolder
        self.unfolder = raw_sol.unfolder

        # Values of activated frontier places in the current Dimacs solution
        # Convenient attribute used:
        # - by FrontierSolution.from_dimacs_front_sol()
        # - to accelerate the call of self.nb_activated_frontiers
        # - as a criterion of uniqueness, to do set operations
        self.activated_frontier_values = raw_sol.extract_activated_frontier_values()

        # Pre-compute nb of activated frontiers in the current solution
        # PS: use numpy to accelerate this ? The cast is costly...
        # We could keep a separated function and use lru_cache instead...
        ## TODO: replace by len() call magic method
        self.nb_activated_frontiers = len(self.activated_frontier_values)

    @classmethod
    def get_unique_dimacs_front_sols_from_raw_sols(cls, raw_sols):
        """Get a tuple of DimacsFrontierSols from a list of RawSolutions.

        The conversion extracts keeps only frontier values from RawSolutions.
        This functions eliminates the duplicates.

        Cf TestMCLAnalyser.py:523:
            raw_sols len 6
            tpl len: 6
        TestMCLAnalyser.py:559
            raw_sols len 12:
            tpl len: 7
        TestMCLAnalyser.py:667
            raw_sols len 7:
            tpl len: 1
        TestMCLAnalyser.py:700
            raw_sols len 7
            tpl len: 1

        ..note:: We need to index the result in MCLAnalyser.less_activated_solution()
            thus, this why we return a tuple here.

        ..note:: The tuple of DimacsFrontierSols is not sorted by order of calculation.
            Thus, the first item is not the first solution found.
            TODO: An order could be used to adjust the number of solutions
            asked during the pruning operation in __solve_with_inact_fsolution()

        :param raw_sols: List of raw solutions;
            most of the time obtained with sq_solutions()
        :return: Set of DimacsFrontierSol
        :type raw_sols: <list <RawSolution>>
        :rtype: <tuple <DimacsFrontierSol>>
        """
        return tuple({DimacsFrontierSol(sol) for sol in raw_sols})

    @property
    def ic_sequence(self):
        """Sequence of activated inputs and events (clocks) in the solution.

        (may be empty if data-flow model is without input)

        .. seealso:: RawSolution.extract_act_input_clock_seq()

        .. note:: Only used in FrontierSolution.from_dimacs_front_sol()
            (at the end of the search process for the cast into FrontierSolution)
            and for tests.
            => This step is expensive !
        """
        # Cf constructor
        return self._ic_sequence_func()

    def get_var_name(self, var_num):
        """For translation to symbolic representation
        Return the string that corresponds to he name of the variable value.
        :rtype: <str>
        """
        return self.unfolder.get_var_name(var_num)

    def nb_timed(self):
        """Count the number of events in the current Dimacs solution (not used)
        :rtype: <int>
        """
        # ic_sequence: list of strings which describes the successive activated
        # inputs and free events
        solution_events = (len(ic_act) for ic_act in self.ic_sequence)
        return sum(it.chain(*solution_events))

    def __len__(self):
        """Get the number of activated frontiers in the current DimacsFrontierSol
        :rtype: <int>
        """
        return len(self.activated_frontier_values)

    def __eq__(self, other):
        """Check if two DimacsFrontierSol have same frontiers values
        Used with __hash__

        ..warning:: We do not test other attributes than activated_frontier_values.
            ic_sequence might be necessary.

        :type other: <DimacsFrontierSol>
        :rtype: <boolean>
        """
        return self.activated_frontier_values == other.activated_frontier_values

    def __ne__(self, other):
        """Python 2 requires this in conjunction to __eq__..."""
        return self.__eq__(other)

    def __hash__(self):
        """Used to test unicity of DimacsFrontierSol in sets;
        Used with __eq__

        ..warning:: We do not test other attributes than activated_frontier_values.
            ic_sequence might be necessary.
        """
        return hash(self.activated_frontier_values)

    def __repr__(self):
        return str(self.activated_frontier_values)

    def __str__(self):
        """For tests and debug"""
        return "<DimacsFrontierSol - Activated frontiers values: %s>".format(self.activated_frontier_values)




# -*- coding: utf-8 -*-
## Filename    : CLUnfolder.py
## Author(s)   : Michel Le Borgne
## Created     : 22 mars 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux IRSET
##
"""Main engine for constraints unfolding and solving.

Process from the intialization with a query to the finding of a solution.


MCLSimpleQuery (reminder):
    Object containing 2 main types of attributes describing properties:

        - Attributes in text format: These are logical formulas that are humanly
        readable.
            Ex: start_prop, inv_prop, final_prop, variant_prop
        - Attributes in DIMACS format: These are logical formulas encoded in
        the form of clauses containing numerical values.
            Ex: dim_start, dim_inv, dim_final, dim_variant_prop

---

About shifting clauses and variables:

    Initialization of the dynamic system by __forward_init_dynamic() iterates on
    all literals of the system.
    If a literal is a t+1 literal, its value is shifted to the right or to the
    left depending on whether its value is positive or negative
    (addition vs soustraction of shift_step value which is the number of
    variables in the system).

    Otherwise, the literal is a t literal and its value stays untouched.

    t+1 and t literals are generated by CLDynSys object directly from the current
    model.

    Shift a clause consists to iterates on all its literals and shift their values
    to the right or to the left depending on whether their values are positive
    or negative.

---

init_with_query(query):
    Copy of the query attributes to temporary attributes internal to the CLUnfolder.

    Textual properties of query are compiled into numeric clauses in
    init_forward_unfolding(), just at the beginning of squery_is_satisfied()
    and squery_solve().

    ALL textual properties are susceptible to add new auxiliary variables in the
    system (increase shift_step). This is why any shift must be made after the
    initialization.


init_forward_unfolding():
    Organization of the initialization of the system (its clauses) for
    the current request and first variables shift.

    Initialized attributes:
        - __initial_constraints:
            Query data + negative values of all places that are not frontiers.
        - __final_constraints:
            Query data
        - __invariant_constraints:
            List of Query data
        - __list_variant_constraints:
            Query data (automatic merge of textual and DIMACS forms)
        - __variant_constraints:
            Initilized with the first item of __list_variant_constraints

    Shifted attributes with __shift_step (nb of variables in the system):
        - __forward_init_dynamic(): __dynamic_constraints:
            Shift dynamic_system.list_clauses + dynamic_system.aux_list_clauses + init
        - __shift_final(): __final_constraints:
            Shift + replace __final_constraints
        - __shift_invariant(): _invariant_constraints:
            Shift + append of the last element of __invariant_constraints

        PS: __variant_constraints si already initialized for the first step


    The following functions decompile textual properties or events and code them
    into DIMACS clauses constraints via __code_clause().
    __code_clause() is susceptible to add new auxiliary numerical variables
    during this process if a variable is not found in __var_code_table and in
    __aux_code_table.

    - __init_initial_constraint_0, __init_final_constraint_0, __init_invariant_constraint_0
        __code_clause + __compile_property
    - __init_variant_constraints_0
        __code_clause + __compile_event


shift():
    Shift all clauses of the system to prepare it for the next step.

    Called at each step by:
        - squery_solve()
        - squery_is_satisfied()

    Modified attributes:
        - __shift_dynamic(): __dynamic_constraints:
            Shift + append of the last element of __dynamic_constraints
        - __shift_variant(): __variant_constraints:
            Shift clauses of the current step in __list_variant_constraints, and
            append them to __variant_constraints.
        - __shift_invariant()
            ...
        - __shift_final()
            ...

        PS: __initial_constraints are left as this for now
        (shifted only with BACKWARD shift direction).

"""
from __future__ import print_function

#from pyCryptoMS import CryptoMS
from pycryptosat import Solver as CryptoMS
from cadbiom.models.biosignal.translators.gt_visitors import compile_cond, compile_event
from cadbiom.models.clause_constraints.CLDynSys import Clause, Literal
from cadbiom.models.clause_constraints.mcl.MCLSolutions import RawSolution, MCLException
from cadbiom import commons as cm

# C++ API
from _cadbiom import shift_clause, shift_dimacs_clauses, forward_code, \
                     forward_init_dynamic

# Standard imports
from logging import DEBUG
import itertools as it

LOGGER = cm.logger()

class CLUnfolder(object):
    """
    When loading a model in a MCLAnalyser object, a CLUnfolder is generated which
    implies the compilation of the dynamical system into a clause constraint
    dynamical system and the DIMACS coding of all the variables.
    This coding cannot be changed later.
    The unfolding of constraints is performed efficiently on the numeric form
    of the clause constraints. The CLUnfolder provide various method to convert
    variable names to DIMACS code and back, to extract the frontier of a model
    and to extract informations from raw solutions.

    Dynamic constraints unfolding management
    ========================================
    Each variable is coded as an integer (in var_code_table and var_list).
    Each clause is represented as a list of signed integers (DIMACS coding).

    During unfolding, clauses are shifted either to the future (forward) or
    to the past (backward). The shift operator is a simple addition
    of self.__shift_step. The shift direction depends only on initializations.
    self.__shift_step depends on the number of variables so it is impossible
    to add variables while generating a trajectory unfolding.

    Glossary
    ========
    - ground variables/ground dimacs code: variables at time 0/their encoding.
    - solution: a solution of the logical dynamic constraint from SAT solver
    - state vector: list of DIMACS code of original (not shifted) variables
      corresponding to a step.
    - trajectory: list of state_vectors


    Some attributes (Please read the constructor comments for much more information):

        __*_property: Logical formulas in text format from the current query.
        __dimacs_*: Clauses in DIMACS format from the current query.
        __*_constraints: CLUnfolder attributes for unfolding and shift initialised
            from the query attributes.

        :param dynamical_system: dynamical system in clause constraint form
        :type dynamical_system: <CLDynSys>
    """

    def __init__(self, dynamic_system, debug=False):
        """
        :param dynamic_system: Describe a dynamic system in clause form.
        :type dynamic_system: <CLDynSys>
        """
        self.dynamic_system = dynamic_system # symbolic clause dynamic system

        # shift_step equals to the total number of coded variables of the system
        # (including inputs, entities, clocks/events, auxiliary variables)

        # shift_step_init: the shift step ss (if n is X_0 code, n+ss is X_1 code)
        self.shift_step_init = dynamic_system.get_var_number()
        self.__shift_step = self.shift_step_init  # current shift/step

        # About unfolder lock and shift_step:
        # self.shift_step must be frozen in order to avoid problems during the
        # unflatten() step of RawSolutions.
        # => Each step MUST have the same number of variables.
        # Thus, we lock the Unfolder by turning self.__locked to True.
        # See __shift_clause() and __m_shift_clause()
        self.__locked = False
        self.__current_step = 1                   # current number of unfolding
        # For reachability optimisation. Number of shifts before checking
        # the final property
        self.__steps_before_check = 0

        # assign a DIMACS code number to each variable (invariant)
        # Create mappings between names and values of the variables of the system
        # dynamic_system.base_var_set : Set of ALL variables of the dynamic system
        # (including inputs, entities, clocks/events, auxiliary variables).
        self.__var_code_table = dict()     # Mapping: name -> DIMACS code (!= 0)
        self.__var_list = ['##']           # Mapping: DIMACS code (!= 0) -> name

        # Fix order of variables names with a list:
        # - indexes of __var_list are the values of the variables at these indexes
        # - values begin from 1 (0 is a blacklisted value)
        if debug:
            base_var_set = sorted(list(dynamic_system.base_var_set))
        else:
            base_var_set = list(dynamic_system.base_var_set)
        self.__var_list += base_var_set
        # keys: are the names of the variables, values: their values
        self.__var_code_table = \
            {var_name: var_num for var_num, var_name in enumerate(base_var_set, 1)}
        # Optimization: for forward_init_dynamic() and forward_code() (mostly
        # implemented in C.
        # Precompute __var_code_table with the literal names in "future" format;
        # i.e with a last char '`' at the end.
        # Thus it is easy to search them directly in the dict (in O(1)),
        # instead of slicing them.
        # Note: "future" literals are already in self.dynamic_system.list_clauses
        # and maybe in self.dynamic_system.aux_list_clauses
        temp_var_code_table = \
            {var_name + "`": var_num for var_name, var_num in self.__var_code_table.iteritems()}
        self.__var_code_table.update(temp_var_code_table)

        assert len(self.__var_code_table) == 2 * self.__shift_step

        # Include auxiliary clause to eliminate undesirable solutions
        # If A->B has clock h, an aux constraint added is h included in h when A
        # Must be set to False for inhibitors computations
        self.__include_aux_clauses = True
        self.__include_aux_clauses_changed = False
        self.__lit_cpt = self.__shift_step + 1    # counter for aux. var. coding

        # Same mapping tools for auxiliary variables introduced by properties
        # compilation
        # Cf __code_clause()
        self.__aux_code_table = dict()            # name to code
        self.__aux_list = []                      # code to name

        # ordered list of DIMACS codes of the simple variables/places
        # Disable all variables of places (entities) in the model (except frontiers)
        self.__no_frontier_init = [[-self.__var_code_table[nfp]] for nfp in self.dynamic_system.no_frontiers]

        # ordered list of DIMACS codes of the frontier variables/places
        self.frontier_values = [self.__var_code_table[frp] for frp in self.dynamic_system.frontiers]
        self.frontier_values.sort()
        ## TODO: sort utile ici ???? si non passer en frozenset et supprimer les casts partout ailleurs
        ## Cf RawSolution.frontier_pos_and_neg_values BACKWARD, encore utilisé avec index
        ## Cf TestCLUnfolder.test_frontier indexable mais peut etre contourné

        # Precompute convenient attributes for:
        # frontiers_pos_and_neg:
        #     - RawSolution.frontier_pos_and_neg_values
        #     (set operation with solution variables)
        # frontiers_negative_values:
        #     - MCLQuery.from_frontier_sol_new_timing
        #     - MCLQuery.frontiers_negative_values
        #     - MCLAnalyser.__solve_with_inact_fsolution
        #     - TestCLUnfolder.test_prune
        #
        # Set of frontier positive and negative values
        # (all frontiers and their opposite version).
        # => operations with sets are much faster
        self.frontiers_negative_values = \
            frozenset(-frontier for frontier in self.frontier_values)
        self.frontiers_pos_and_neg = \
            self.frontiers_negative_values | frozenset(self.frontier_values)

        # DIMACS codes of the input variables/places for extraction (inv)
        self.__inputs = frozenset(self.__var_code_table[inp] for inp in self.dynamic_system.inputs)

        # DIMACS codes of the free_clocks variables/places for extraction (invariant)
        self.__free_clocks = frozenset(self.__var_code_table[fcl] for fcl in self.dynamic_system.free_clocks)

        # Binding for a merged version of __inputs and __free_clocks
        # Convenient attribute for RawSolution.extract_act_input_clock_seq()
        # DIMACS codes of the input and free_clocks variables
        self.inputs_and_free_clocks = self.__inputs | self.__free_clocks

        # Properties to be checked
        self.reset()

        # Logical constraints:
        # Result from unfolding of base constraints
        # Boolean vectors signification:
        # X: Current state of places (activated/unactivated)
        # X': Future state of places
        # H: Current 'free' events (present/not present) => ?
        # I: Current inputs => ?
        self.__dynamic_constraints = []        # DIMACS clauses: X' = f(X,H,I)
        self.__precomputed_dynamic_constraints = []

        # Simple temporal properties:
        # SP(X0): Initial property/start property; Never change at each step.
        # IP(X): Invariant property
        # VP(X): Variant property
        #   List of logical formulas of properties forced at each step
        #   It's the trajectory of events of a solution.
        # FP(X): Final property

        self.__initial_constraints = []        # DIMACS clauses: C(X_0) <list <list <int>>>
        self.__final_constraints   = []        # idem: C(X_n)
        self.__invariant_constraints = []      # DIMACS clauses: C(X_i))
        self.__variant_constraints = []        # variant constraints along trajectory
        self.__shift_direction = None          # FORWARD or BACKWARD

        # statistics on solver
        self.__stats = False
        self.__nb_vars = 0
        self.__nb_clauses = 0


    def reset(self):
        """Reset the unfolder before a new query

        Reset only properties and dimacs clauses from the current query;
        AND __list_variant_constraints.

        => __initial_constraints, __final_constraints, __invariant_constraints,
        __variant_constraints and __dynamic_constraints ARE NOT reset here
        (see init_forward_unfolding())

        This function is called from the constructor and during
        MCLAnalyser.sq_is_satisfiable() and MCLAnalyser.sq_solutions()
        following the call of init_with_query().
        """
        # Properties to be checked
        self.__initial_property = None            # logical formula - literal boolean expression
        self.__dimacs_initial = None              # list of DIMACS clauses
        self.__final_property = None              # logical formula
        self.__dimacs_final = None                # list of DIMACS clauses
        self.__invariant_property = None          # logical formula
        self.__dimacs_invariant = None            # list of DIMACS clauses
        self.__variant_property = None            # list<logic formulas>
        self.__dimacs_variant = None              # list<list<DIMACS clauses>>
        # list of variant temporal constraints in Dimacs ground code
        self.__list_variant_constraints = None    # list<list<DIMACS clauses>>

        # If this function is called from the constructor,
        # the following variables are just redefined here.

        # For reachability optimisation. Number of shifts before checking
        # the final property
        self.__steps_before_check = 0
        self.__shift_direction = None             # FORWARD or BACKWARD
        self.__locked = False

    def init_with_query(self, query):
        """Initialise the unfolder with the given query

        Following attributes are used from the query:
            start_prop, dim_start, inv_prop, dim_inv, final_prop, dim_final,
            variant_prop, dim_variant_prop, steps_before_check

        Textual properties of query are compiled into numeric clauses in
        init_forward_unfolding(), just at the beginning of squery_is_satisfied()
        and squery_solve().
        This compilation step is costly due to ANTLR performances...

        .. warning:: ALL textual properties are susceptible to add new auxiliary
            variables in the system (increase shift_step).
        """
        # Reset the unfolder before a new query
        self.reset()
        # Init with query properties and clauses
        self.__initial_property = query.start_prop # logical formula in text
        self.__dimacs_initial = query.dim_start

        self.__final_property = query.final_prop # logical formula in text
        self.__dimacs_final = query.dim_final

        self.__invariant_property = query.inv_prop # logical formula in text
        self.__dimacs_invariant = query.dim_inv

        # It's the trajectory of events of a solution.
        self.__variant_property = query.variant_prop # logical formula in text
        self.__dimacs_variant = query.dim_variant_prop

        # For reachability optimisation. Number of shifts before checking
        # the final property (default 0)
        self.__steps_before_check = query.steps_before_check

    def set_stats(self):
        """Enable solver statistics (for tests)"""
        self.__stats = True
        self.__nb_vars = 0
        self.__nb_clauses = 0

    def unset_stats(self):
        """Disable solver statistics (never used)"""
        self.__stats = False

    def _stats(self):
        """Display solver statistics"""
        print("\n NB Variables in solver:", self.__nb_vars)
        print("NB Clauses in solver:", self.__nb_clauses)

    def set_include_aux_clauses(self, val):
        """Flag to include auxiliary clause to eliminate undesirable solutions.

        If A->B has clock h, an aux constraint added is h included in h when A
        Must be set to False for inhibitors computations.

        .. warning:: This function impacts the shift_step number.

        Tested in by:
            - __forward_init_dynamic
            - __backward_init_dynamic
        """
        self.__include_aux_clauses = val
        self.__include_aux_clauses_changed = True

    ## Internal variable access for tests (never used) #########################
    @property
    def dynamic_constraints(self):
        """For tests: returns coded dynamic constraints"""
        return self.__dynamic_constraints

    @property
    def initial_constraints(self):
        """For tests: returns coded initial constraints"""
        return self.__initial_constraints

    @property
    def invariant_constraints(self):
        """For tests: returns coded invariant constraints"""
        return self.__invariant_constraints

    @property
    def variant_constraints(self):
        """For tests: returns coded variant constraints"""
        return self.__variant_constraints

    @property
    def final_constraints(self):
        """For tests: returns coded final constraints"""
        return self.__final_constraints

    ## Variables management ####################################################
    def var_names_in_clause(self, clause):
        """Return the names of the variables from values in the given numeric clause
        (DEBUG never used)
        """
        return [self.get_var_indexed_name(var) for var in clause]

    def var_dimacs_code(self, var_name):
        """Returns DIMACS code of var_name (string) variable (for tests)"""
        return self.__var_code_table[var_name]

    def get_system_var_number(self):
        """Get number of variables in the clause constraint dynamical system
        (including inputs, entities, clocks/events, auxiliary variables) (for tests)

        .. note:: This number should be equal to the number of variables
            in self.get_var_number()
        """
        return self.dynamic_system.get_var_number()

    def get_var_number(self):
        """Get number of principal variables (properties excluded) in the unfolder
        (including inputs, entities, clocks/events, auxiliary variables) (for tests)

        .. note:: This number should be equal to the number of variables
            in self.get_system_var_number()
        """
        # Remove the blacklisted first item (##) for variables naming
        return len(self.__var_list) - 1

    def get_var_name(self, var_num):
        """Get name of the variable

        .. seealso:: Explanations on get_var_indexed_name()

        @param var_num: DIMACS literal coding of an initial variable
        @return: name of the variable
        """
        # Get the original var_code without shifts
        var_code = (abs(var_num) - 1) % self.__shift_step + 1
        if var_code <= self.shift_step_init:
            # Variable num is less than the number of the variables in the system
            # => variable from the initial system
            return self.__var_list[var_code]
        else:
            # auxiliary variable introduced by properties compilation
            return self.__aux_list[var_code - self.shift_step_init -1]
            #raise MCLException("Not a DIMACS code of an initial variable")

    def get_var_indexed_name(self, var_num):
        """Get name of the variable with the time index appended

        .. note:: time index is the number of steps since the begining of the
            simulation.

            :Example:
                __shift_step = 2 (current step of the unfolding)
                shift_step_init = 2 (number of variables in the system)
                __var_list = ["##", "n1", "n2"]

                Virtual list of indexes in the state vector of a solution:
                [0, 1, 2, 3, 4, 5, 6, 7, 8, ...]
                 #|n1,n2|n1,n2|n1,n2|n1,n2|
                   ti=0  ti=1  ti=2  ti=3

                Given var_num = 7:
                index of var_name in __var_list: ((7-1) % 2) + 1 = 1
                var_name = "n1"
                time index: (7 - 1) / 2 = 3

                => return: "n1_3"

        @param var_num: DIMACS literal coding
        @return: name of the variable with the time index appended
        """
        varnum1 = abs(var_num)
        # Get the original var_code without shifts
        var_code = (varnum1 - 1) % self.__shift_step + 1
        var_step = (varnum1 - var_code) / self.__shift_step
        if var_code <= self.shift_step_init:
            # Variable num is less than the number of the variables in the system
            # => variable from the initial system
            return self.__var_list[var_code] + '_%s'% var_step
        else:
            # Auxiliary variable introduced by properties compilation
            index = var_code - self.shift_step_init - 1
            return self.__aux_list[index] + '_%s'% var_step

    def get_free_clocks(self):
        """Get the DIMACS codes of the free_clocks variables"""
        return self.__free_clocks

    def get_inputs(self):
        """Get the DIMACS codes of the input variables"""
        return self.__inputs

    def get_shift_direction(self):
        """
        @return: string "FORWARD" or "BACKWARD"
        """
        return self.__shift_direction

    def get_shift_step(self):
        """
        @return: the shift step ss (if n is X_0 code, n+ss is X_1 code)
        """
        return self.__shift_step

    def get_current_step(self):
        """
        @return: the current number of unfolding
        """
        return self.__current_step


    ## Translation from names to num codes #####################################
    ## The translation depends on the shift direction
    def __forward_code(self, clause):
        """(deprecated, directly included in C++ module: forward_init_dynamic())

        Numerically code a clause with the numeric code found in
        self.__var_code_table for a base variable x,
        and numeric_code + shift_step for x' variable

        ..note:: Future variables x' are noted "name`" in Literal names.
            Values of variables increases of __shift_step in future steps.

        @param clause: a Clause object
        @return: the DIMACS coding of the forward shifted clause
        """
        # Old API
        # num_clause = []
        # for lit in clause.literals:
        #     if lit.name[-1] == '`': # t+1 variable
        #         num_clause.append(
        #             -(self.__var_code_table[lit.name[:-1]] + self.__shift_step) \
        #             if not lit.sign \
        #             else (self.__var_code_table[lit.name[:-1]] + self.__shift_step)
        #         )
        #     else: # t variable
        #         num_clause.append(
        #             -self.__var_code_table[lit.name] \
        #             if not lit.sign else self.__var_code_table[lit.name]
        #         )
        # return num_clause

        # New API via C++ module
        return forward_code(clause, self.__var_code_table, self.__shift_step)


    def __backward_code(self, clause):
        """(never used, backward is partially implemented)

        numerically code a clause with the numeric code found in
        self.__var_code_table + shift_step for a base variable x,
        and numeric_code for x' variable integer coding increases in past steps

        @param clause: a Clause object
        @return: the DIMACS coding of the backward shifted clause
        """
        num_clause = []
        for lit in clause.literals:
            if lit.name[-1] == '`': # t+1 variable
                num_clause.append(
                    -self.__var_code_table[lit.name[:-1]] \
                    if not lit.sign \
                    else self.__var_code_table[lit.name[:-1]]
                )
            else: # t variable, base variable
                num_clause.append(
                    -(self.__var_code_table[lit.name] + self.__shift_step) \
                    if not lit.sign \
                    else (self.__var_code_table[lit.name] + self.__shift_step)
                )
        return num_clause

    def __code_clause(self, clause):
        """Numerically code a clause

        The numerical values are found in
        self.__var_code_table for variables in the dynamic system,
        or in self.__aux_code_table for other auxiliary variables;
        assume all variables are basic variables (t=0) (not shifted).

        @warning: This function MODIFIES SHIFT_STEP because we add a variable
            in the system if a variable is not found in __var_code_table or in
            __aux_code_table.
            A new auxiliary variable will be created in __aux_code_table.

        :param clause:
        :return: List of numerical values corresponding to the literals
            in the given clause.
        :type clause: <Clause>
        :rtype: <list <int>>
        """
        num_clause = []
        for lit in clause.literals:
            # Get variable value with the given name
            name = lit.name
            if name in self.__var_code_table:
                var_cod = self.__var_code_table[name]
            elif name in self.__aux_code_table:
                var_cod = self.__aux_code_table[name]
            else:
                # Create an auxiliary variable - modifies shift_step (nb of variables)
                self.__shift_step += 1
                # Mapping name to value code
                self.__aux_code_table[name] = self.__shift_step
                # Mapping value code to name
                self.__aux_list.append(name)
                # Define var_cod
                var_cod = self.__shift_step

            # Add the sign to the value
            lit_code = var_cod if lit.sign else -var_cod

            num_clause.append(lit_code)
        return num_clause

    ## Dynamic initialisations #################################################
    def __forward_init_dynamic(self):
        """Dynamics initialisations.
        Set dynamic constraints for a forward one step: X1 = f(X0)

        Numerically code clauses with the numeric codes found in
        self.__var_code_table for a base variable x,
        and numeric_code + shift_step for x' variable integer coding increases
        in future steps.

        __dynamic_constraints is a list of lists of numeric clauses (lists of ints)
        Each sublist of __dynamic_constraints corresponds to a step in the unfolder;
        the last step is the last element.

        .. seealso:: __forward_code()

        .. note:: Called by init_forward_unfolding()

        .. note:: Future variables x' are noted "name`" in Literal names.

        .. note:: Values of variables increases of __shift_step in future steps.
        """
        # New API via C++ module
        # TODO: take a look at __backward_init_dynamic & __backward_code
        if self.__include_aux_clauses:
            # Auxiliary clauses are supported (default)
            self.__dynamic_constraints = \
                forward_init_dynamic(self.dynamic_system.list_clauses,
                                     self.__var_code_table,
                                     self.__shift_step,
                                     self.dynamic_system.aux_list_clauses)
        else:
            self.__dynamic_constraints = \
                forward_init_dynamic(self.dynamic_system.list_clauses,
                                     self.__var_code_table,
                                     self.__shift_step)

        # __dynamic_constraints is initialized and
        # has now taken into account the auxiliary clauses flag
        self.__include_aux_clauses_changed = False

    def __backward_init_dynamic(self):
        """Dynamics initialisations. (never used, backward is partially implemented)
        Set dynamic constraints for a forward one step: X0 = f(X1)

        .. note:: Called by init_backward_unfolding()
        """
        num_clause_list = \
            [self.__backward_code(clause)
             for clause in self.dynamic_system.list_clauses]

        if self.__include_aux_clauses:
            num_clause_list += \
                [self.__backward_code(clause)
                 for clause in self.dynamic_system.aux_list_clauses]

        self.__dynamic_constraints = [num_clause_list]

    ## Shifting variables: implementation of the shift operator ################
    def __shift_clause(self, numeric_clause):
        """Shift a clause for the current __shift_step

        (no used anymore, replaced by direct use of C++ code)

        Basically, `shift_step` is added to positive variables and subtracted
        from negative variables in `numeric_clause`.

        About unfolder lock and shift_step:
        self.shift_step must be frozen in order to avoid problems during the
        unflatten() step of RawSolutions.
        => Each step MUST have the same number of variables.
        Thus, we lock the Unfolder by turning self.__locked to True.

        @param numeric_clause: DIMACS clause
        @warning: lock the unfolder
        """
        # Froze unfolder to avoid modifications of shift_step
        self.__locked = True

        # Old API
        # Less efficient with abs()
        # return [(abs(lit) + self.__shift_step) * (-1 if lit < 0 else 1)
        #         for lit in numeric_clause]
        # More efficient with ternary assignment
        # return [(lit + self.__shift_step) if lit > 0 else (lit - self.__shift_step)
        #         for lit in numeric_clause]

        # New API via C++ module
        return shift_clause(numeric_clause, self.__shift_step)

    def __m_shift_clause(self, numeric_clause, nb_steps):
        """Shift a clause for the given step number

        Each literal will be shifted by (nb_steps * __shift_step).

        Why ?
        Such function is used for clauses that are not already shifted during
        previous steps. It is the case of variant_constraints which come from a
        reloaded/forced trajectory.

        About unfolder lock and shift_step:
        self.shift_step must be frozen in order to avoid problems during the
        unflatten() step of RawSolutions.
        => Each step MUST have the same number of variables.
        Thus, we lock the Unfolder by turning self.__locked to True.

        Called by: __shift_variant()

        @param numeric_clause: DIMACS clause
        @param nb_steps: number of shifts asked
        @warning: lock the unfolder
        """
        # Froze unfolder to avoid modifications of shift_step
        self.__locked = True

        return [(lit + self.__shift_step * nb_steps) if lit > 0
                else (lit - self.__shift_step * nb_steps) for lit in numeric_clause]

    def __shift_dynamic(self):
        """Shift clauses representing the dynamics X' = f(X,I,C)

        Shift the last item of self.__dynamic_constraints and append the result
        to self.__dynamic_constraints.

        .. note:: Called by shift()
        """

        if len(self.__precomputed_dynamic_constraints) > self.__current_step:
            # Recall the saved state of the system for the next step
            self.__dynamic_constraints.append(
                self.__precomputed_dynamic_constraints[self.__current_step]
            )
            LOGGER.info(
                "shift_dynamic:: OPTI DO NOT SHIFT; "
                "Reload dynamic constraints; step: %s",
                self.__current_step
            )
            return

        # System has not been shifted for this step until now
        LOGGER.info(
                "shift_dynamic:: SHIFT; "
                "Set dynamic constraints; step: %s; dyn consts len: %s",
                self.__current_step,
                len(self.__dynamic_constraints)
            )

        # Old API
        # self.__dynamic_constraints.append(
        #   [self.__shift_clause(clause)
        #    for clause in self.__dynamic_constraints[-1]]
        # )

        # New API via C++ module
        # __dynamic_constraints:
        # List of lists of DIMACS encoded clauses (lists of ints)
        self.__dynamic_constraints.append(
            shift_dimacs_clauses(
                self.__dynamic_constraints[-1], # <list <list <int>>>
                self.__shift_step
            )
        )

    def __shift_initial(self):
        """Shift initialisation condition

        Shift + replace __initial_constraints

        .. note:: Called by:
            - shift()
            - init_backward_unfolding()
        """
        self.__initial_constraints = \
            shift_dimacs_clauses(
                self.__initial_constraints,
                self.__shift_step
            )

    def __shift_final(self):
        """Shift final property

        Shift + replace __final_constraints

        .. note:: Called by:
            - shift()
            - init_forward_unfolding()
        """
        self.__final_constraints = \
            shift_dimacs_clauses(
                self.__final_constraints,
                self.__shift_step
            )

    def __shift_invariant(self):
        """Shift invariant property

        Shift + append of the last element of __invariant_constraints

        .. note:: Called by:
            - shift()
            - init_forward_unfolding()
            - init_backward_unfolding()
        """
        if self.__invariant_constraints:
            # New API via C++ module
            self.__invariant_constraints.append(
                shift_dimacs_clauses(
                    self.__invariant_constraints[-1],
                    self.__shift_step
                )
            )

    def __shift_variant(self):
        """Shift variant property - depends on unfolding direction

        Shift clauses of the current step in __list_variant_constraints, and
        append them to __variant_constraints.

        .. warning:: Refactor note: unclear function. No apparent reason for
            only the first clause to be shifted...

        Process:
            - Take the clauses (constraint) of the current step
            - Shift the first clause
            - Append it to __variant_constraints
            - Stop ????

            Example:
            current_step: 2 (shift for step 3)
            [[], [[4], [4]], [[6], [7]], []]
            => Only [6] is shifted...

        .. note:: Called by shift()
        """
        if not self.__list_variant_constraints:
            return
        if self.__shift_direction == "FORWARD":
            # Constraint t0 already included from the query during initialisation
            current_constraint = self.__list_variant_constraints[self.__current_step]
            # Shift constraints at current step and append them to __variant_constraints
            for clause in current_constraint:
                s_clause = self.__m_shift_clause(clause, self.__current_step)
                self.__variant_constraints.append(s_clause)
                return # ?????
        elif self.__shift_direction == "BACKWARD":
            raise MCLException("Not yet implemented")
        else:
            raise MCLException("Shift incoherent data: " + self.__shift_direction)


    def shift(self):
        """Shift all clauses of the system to prepare it for the next step.

        self.__current_step is incremented here!

        Modified attributes:
            - __dynamic_constraints[-1] (last element)
                Shift + append of the last element of __dynamic_constraints
            - __invariant_constraints[-1] (last element)
                Shift + append of the last element of __invariant_constraints
            - __list_variant_constraints[current_step]
                Shift clauses of the current step in __list_variant_constraints,
                and append them to __variant_constraints.
            - __final_constraints if __shift_direction is "FORWARD"
            - __initial_constraints if __shift_direction is "BACKWARD"

            PS: __initial_constraints are left as this for now
            (shifted only with BACKWARD shift direction)

        Called at each step by:
            - squery_solve()
            - squery_is_satisfied()
        """
        # Froze unfolder to avoid modifications of shift_step
        self.__locked = True

        self.__shift_dynamic()
        self.__shift_invariant()
        self.__shift_variant()
        if self.__shift_direction == 'FORWARD':
            self.__shift_final()
        elif self.__shift_direction == 'BACKWARD':
            self.__shift_initial()
        else:
            # Shift direction must be set
            raise MCLException("Shift incoherent data: " + self.__shift_direction)

        # Increment the current step
        self.__current_step += 1

    ## Coding of properties ####################################################
    def __compile_property(self, property_text):
        """Compile a property (logical formula) into clauses

        Type checking uses the symbol table of dyn_sys
        Error reporting is made through the dyn_sys reporter

        .. warning:: MODIFIES __lit_cpt for numbering of auxiliary variables (not coding)

        .. note:: Called by:
            - __init_initial_constraint_0()
            - __init_final_constraint_0()
            - __init_invariant_constraint_0()
        """
        if self.__locked:
            raise MCLException("Trying to compile property while unfolder is locked")

        # Syntax analyser and type checker
        # Compile a condition expression to a tree representation
        tree_prop = compile_cond(
            property_text,
            self.dynamic_system.symb_tab,
            self.dynamic_system.report
        )
        # Avoid name collisions of aux var
        prop_visitor = CLPropertyVisitor(self.__lit_cpt)
        tree_prop.accept(prop_visitor)
        self.__lit_cpt = prop_visitor.cpt # avoid name collisions of aux var
        return prop_visitor.clauses

    def __compile_event(self, property_text):
        """Compile an event (biosignal expression) into clauses

        Type checking uses the symbol table of dyn_sys
        Error reporting is made through the dyn_sys reporter

        .. warning:: MODIFIES __lit_cpt for numbering of auxiliary variables (not coding)

        .. note:: Called by:
            - __init_variant_constraints_0()
        """
        if self.__locked:
            raise MCLException("Trying to compile property while unfolder is locked")

        # Syntax analyser and type checker
        # Compile an event expression to a tree representation
        # tree_prop is the event expression,
        # ste the state events (s#> ..) used and
        # fcl is the free clocks used in the event expression.
        tree_prop, ste, fcl = compile_event(
            property_text,
            self.dynamic_system.symb_tab,
            True, # Collect free clocks (events)
            self.dynamic_system.report
        )
        # Avoid name collisions of aux var
        prop_visitor = CLPropertyVisitor(self.__lit_cpt)
        tree_prop.accept(prop_visitor)
        self.__lit_cpt = prop_visitor.cpt # avoid name collisions of aux var
        return prop_visitor.clauses

    def __init_initial_constraint_0(self, no_frontier_init=True):
        """Code initial constraints in a numerical clause.

        If initial property is set (in text format of logical formulas, or
        in dimacs format of numerical values), this function generates
        constraints clauses in numerical form.

        Initialisation of self.__initial_constraints:
        List of DIMACS clauses (lists of values).

        self.__initial_constraints is filled with the current query data AND
        with a clause with ALL places that are not frontiers (negative values).

        We have to wait until all variables are num coded before shifting anything!

        .. note:: Compilation of properties in text format (logical formulas)
            to numeric format is expensive when they must be parsed at each query.

        :key no_frontier_init: (optional) boolean to force all variables of
            places/entities that are not frontiers to be disabled before each
            search.
            default: True
        :type no_frontier_init: <boolean>
        """
        self.__initial_constraints = list()

        if no_frontier_init:
            # Disable all variables of places (entities) in the model (except frontiers)
            self.__initial_constraints += self.__no_frontier_init

        if self.__initial_property:
            # Compile initial property from text logical formulas to numeric form
            # __initial_property: is taken from the current query
            # Costly...
            self.__initial_constraints += \
                [self.__code_clause(clause) for clause
                 in self.__compile_property(self.__initial_property)]

        if self.__dimacs_initial:
            # Add DIMACS aux clauses initial properties
            self.__initial_constraints += self.__dimacs_initial

    def __init_final_constraint_0(self):
        """Code final constraints in a numerical clause.

        If final property is set (in text format of logical formulas, or
        in dimacs format of numerical values), this function generates
        constraints clauses in numerical form.

        Initialisation of self.__final_constraints:
        List of DIMACS clauses (lists of values).

        We have to wait until all variables are num coded before shifting anything!

        .. note:: Compilation of properties in text format (logical formulas)
            to numeric format is expensive when they must be parsed at each query.
        """
        self.__final_constraints = list()

        if self.__final_property:
            # compile initial (X0) property into numeric form
            # Ex: [$Px$] => self.__final_constraints = [[7]]
            self.__final_constraints += \
                [self.__code_clause(clause) for clause
                 in self.__compile_property(self.__final_property)]

        if self.__dimacs_final:
            # Add DIMACS aux clauses initial properties
            self.__final_constraints += self.__dimacs_final

    def __init_invariant_constraint_0(self):
        """Code final constraints in a numerical clause.

        If trajectory property is set (in text format of logical formulas, or
        in dimacs format of numerical values), this function generates
        constraints clauses in numerical form.

        Initialisation of self.__invariant_constraints:
        List of lists of DIMACS clauses (lists of values).

        We have to wait until all variables are num coded before shifting anything!

        .. note:: Compilation of properties in text format (logical formulas)
            to numeric format is expensive when they must be parsed at each query.
        """
        self.__invariant_constraints = list()

        if self.__invariant_property:
            # compile initial (X0) property into numeric form
            self.__invariant_constraints.append(
                [self.__code_clause(clause) for clause
                 in self.__compile_property(self.__invariant_property)]
            )

        if self.__dimacs_invariant:
            # Add DIMACS aux clauses initial properties
            self.__invariant_constraints.append(self.__dimacs_invariant)

    def __init_variant_constraints_0(self):
        """Code variant constraints in a numerical clause.

        .. TODO:: variant_property list d'étapes, chaque étape a une formule logique
            impliquant les évènements (et des places? pas sur...
            Cf __compile_event())
            => ensemble de place/propriétés devant etre valides à chaque étape

        Automatic merge of textual and DIMACS forms:
            If variant_property is set, compile each property into clauses and
            encode these clauses. Clauses already in DIMACS form are added.
            The two variant constraint forms must be compatible (same length) in
            order to allow a merge.

        self.__list_variant_constraints:
            Built here and is used later to shift current variant properties
            and update self.__variant_constraints in __shift_variant().

        self.__variant_constraints:
            - the only element passed to the solver.
            - initilised with the first item of self.__list_variant_constraints
        """
        # coding of variant properties
        # Take __variant_property in priority on __dimacs_variant
        if self.__variant_property:
            self.__list_variant_constraints = list()

            for prop in self.__variant_property:
                # compile initial (X0) property into numeric form
                # For each property in step of __variant_property
                self.__list_variant_constraints.append(
                    [self.__code_clause(clause)
                     for clause in self.__compile_event(prop)]
                )
                print(self.__list_variant_constraints)

            if self.__dimacs_variant:
                # Add DIMACS aux clauses initial properties
                # Check if the number of steps is equal
                if len(self.__variant_property) != len(self.__dimacs_variant):
                    raise MCLException(
                        "Incoherent variant properties; sizes of "
                        "__variant_property and __dimacs_variant differ"
                    )

                # Merge together all steps content
                # Ex:
                # __list_variant_constraints due to __variant_property convertion:
                # [[], [[4]], [], []]
                # __dimacs_variant:
                # [[], [[4]], [[6], [7]], []]
                # Result:
                # [[], [[4], [4]], [[6], [7]], []]
                self.__list_variant_constraints = \
                    [list(it.chain(*cpl)) for cpl
                     in zip(self.__dimacs_variant, self.__list_variant_constraints)]

        elif self.__dimacs_variant:
            # No __variant_property, keep only __dimacs_variant
            self.__list_variant_constraints = self.__dimacs_variant

        # Initialisation
        if self.__list_variant_constraints:
            if self.__shift_direction == "FORWARD":
                # For now, keep only the events of the first step (t0)
                # Other steps are taken during shift() calls
                self.__variant_constraints = self.__list_variant_constraints[0]
            elif  self.__shift_direction == "BACKWARD":
                raise MCLException("Not yet implemented")
            else:
                raise MCLException("Shift incoherent data: " + self.__shift_direction)

    ## Whole initialisations ###################################################
    def init_forward_unfolding(self):
        """Initialisation before generating constraints - forward trajectory

        Organization of the initialization of the system (its clauses) for
        the current request and first variables shift.

        .. warning:: ALL textual properties are susceptible to add new auxiliary
            variables in the system (increase shift_step).
            This is why the shift is made after initialization.

        Initialized attributes:
            - __initial_constraints:
                Query data + negative values of all places that are not frontiers.
            - __final_constraints:
                Query data
            - __invariant_constraints:
                List of Query data
            - __list_variant_constraints:
                Query data (automatic merge of textual and DIMACS forms)
            - __variant_constraints:
                Initilized with the first item of __list_variant_constraints

        Shifted attributes with __shift_step (nb of variables in the system):
            - __dynamic_constraints:
                Shift dynamic_system.list_clauses + dynamic_system.aux_list_clauses + init
            - __final_constraints:
                Shift + replace __final_constraints
            - __invariant_constraints:
                Shift + append of the last element of __invariant_constraints

            PS: __variant_constraints si already initialized for the first step.


        The following functions use the following methods to decompile textual
        properties or events (__compile_property(), __compile_event()) and to
        code them into DIMACS clauses constraints (__code_clause()).

        __code_clause() is susceptible to add new auxiliary numeric variables
        during this process if a variable is not found in __var_code_table and in
        __aux_code_table.

        - __init_initial_constraint_0, __init_final_constraint_0, __init_invariant_constraint_0
            __code_clause + __compile_property
        - __init_variant_constraints_0
            __code_clause + __compile_event

        Called at the begining of squery_is_satisfied() and squery_solve()
        """
        old_shift_step = self.__shift_step
        self.__shift_direction = 'FORWARD'
        self.__current_step = 1
        self.__shift_step = self.shift_step_init  # back to basic!
        self.__aux_code_table = dict()            # flush auxiliary variables
        self.__aux_list = []                      # idem

        # Init properties to generate all variable num codes
        self.__init_initial_constraint_0()
        self.__init_final_constraint_0()
        self.__init_invariant_constraint_0()
        self.__init_variant_constraints_0()

        # Now shifting is possible
        if old_shift_step != self.__shift_step \
        or self.__include_aux_clauses_changed \
        or not self.__dynamic_constraints:
            # The size of the system has changed,
            # or aux_clauses flag has been updated (so the system has changed),
            # or the system has never been initialized before.
            LOGGER.info("init_forward_unfolding:: New dynamic constraints")
            self.__forward_init_dynamic()
            self.__precomputed_dynamic_constraints = []
        else:
            # Main settings of the unfolder haven't changed and the system
            # was initialized before.
            LOGGER.info("init_forward_unfolding:: Reload dynamic constraints")
            # Save the whole system
            self.__precomputed_dynamic_constraints = self.__dynamic_constraints
            # Recall the initial state of the system
            self.__dynamic_constraints = [self.__dynamic_constraints[0]]
        self.__shift_final()
        self.__shift_invariant()


    def init_backward_unfolding(self):
        """Initialisation before generating constraints - backward trajectory
        Never used (and backward not implemented in __init_variant_constraints_0)
        """
        self.__shift_direction = 'BACKWARD'
        self.__current_step = 0
        self.__shift_step = self.shift_step_init  # back to basic!
        self.__aux_code_table = dict()            # flush auxiliary variables
        self.__aux_list = []                      # idem

        # Init properties to generate all variable num codes
        self.__init_initial_constraint_0()
        self.__init_final_constraint_0()
        self.__init_invariant_constraint_0()
        self.__init_variant_constraints_0()

        # Now shifting is possible
        self.__backward_init_dynamic()
        self.__shift_initial()
        self.__shift_invariant()

    ## Solver interface ########################################################
    def __load_solver(self, solv):
        """Add all the current clauses in the solver

        .. note:: Called by:
            - __constraints_satisfied()
            - __msolve_constraints()
        """
        # New API via C++ module
        solv.add_clauses(self.__final_constraints)
        solv.add_clauses(it.chain(*self.__invariant_constraints))
        solv.add_clauses(self.__variant_constraints)
        solv.add_clauses(it.chain(*self.__dynamic_constraints))
        solv.add_clauses(self.__initial_constraints)

        if LOGGER.getEffectiveLevel() == DEBUG:
            # final
            LOGGER.debug("Load new solver !!")
            LOGGER.debug(">> final constraints: %s", len(self.__final_constraints))
            LOGGER.debug(str(self.__final_constraints))

            # trajectory invariant
            LOGGER.debug(">> trajectory invariant constraints: %s", len(self.__invariant_constraints))
            LOGGER.debug(str(self.__invariant_constraints))

            # trajectory variant
            LOGGER.debug(">> trajectory variant constraints: %s", len(self.__variant_constraints))
            LOGGER.debug(str(self.__variant_constraints))

            # dynamics
            LOGGER.debug(">> dynamic constraints: %s", len(self.__dynamic_constraints))
            LOGGER.debug(str(self.__dynamic_constraints))

            # initial
            LOGGER.debug(">> initial constraints: %s", len(self.__initial_constraints))
            LOGGER.debug(str(self.__initial_constraints))


    def __constraints_satisfied(self):
        """Ask the SAT solver if the query/system is satisfiable in the given
        number of steps.

        .. warning:: vvars (frontier values) are not tested. Thus the satisfiability
            is not tested for interesting variables like in squery_solve().

        :return: True if the problem is satisfiable, False otherwise.
        :rtype: <boolean>
        """
        solver = CryptoMS()
        # Load all clauses into the solver
        self.__load_solver(solver)

        if self.__stats:
            # If stats are activated (never besides in test environment):
            # sync local data (nb vars, nb clauses with solver state)
            if solver.nb_vars() > self.__nb_vars:
                self.__nb_vars = solver.nb_vars()

            # Starting from pycryptosat 5.2, nb_clauses() is a private attribute
            #if solver.nb_clauses() > self.__nb_clauses:
            #    self.__nb_clauses = solver.nb_clauses()
        # Is the problem satisfiable ?
        return solver.is_satisfiable()


    def __msolve_constraints(self, max_sol, vvars):
        """Get solutions from the solver

        :param max_sol: The maximum number of solutions to be returned
        :param vvars: Variables for which solutions must differ.
            In practice, it is a list with the values (Dimacs code)
            of all frontier places of the system.
        :type max_sol: <int>
        :type vvars: <list <int>>
        :return: A tuple of RawSolution objects
            RawSolution objects contain a solution got from SAT solver with all
            variable parameters from the unfolder
        :rtype: <tuple <RawSolution>>
        """
        solver = CryptoMS()
        # Load all clauses into the solver
        self.__load_solver(solver)

        if self.__stats:
            # If stats are activated (never besides in test environment):
            # sync local data (nb vars, nb clauses with solver state)
            if solver.nb_vars() > self.__nb_vars:
                self.__nb_vars = solver.nb_vars()
            # Starting from pycryptosat 5.2, nb_clauses() is a private attribute
            #if solver.nb_clauses() > self.__nb_clauses:
            #    self.__nb_clauses = solver.nb_clauses()

        if LOGGER.getEffectiveLevel() == DEBUG:
            LOGGER.debug("__msolve_constraints :: vvars : %s", vvars)
            LOGGER.debug("__msolve_constraints :: max_sol : %s", max_sol)
            # Get List of solutions (list of tuples of literals)
            lintsol = solver.msolve_selected(max_sol, vvars)
            LOGGER.debug("__msolve_constraints :: lintsol : %s", lintsol)
            # Make a RawSolution for each solution (tuple of literals)
            return [RawSolution(solint, self) for solint in lintsol]

        return tuple(RawSolution(solint, self)
                for solint in solver.msolve_selected(max_sol, vvars))

    def squery_is_satisfied(self, max_step):
        """Ask the SAT solver if the query/system is satisfiable in the given
        number of steps.

        If __invariant_property is set and __final_property is not set,
        1 satisfiability test is made only after all shift() calls for each
        remaining step from the current step.

        Otherwise, 1 test is made for each remaining step after each shift() call.

        .. warning:: vvars (frontier values) are not tested. Thus the satisfiability
            is not tested for interesting variables like in squery_solve().

        .. note:: This function checks and corrects incoherent data on step number.
            __steps_before_check can't be >= max_step

        :param max_step: The horizon on which the properties must be satisfied.
        :type max_step: <int>
        """
        # Initialization
        self.init_forward_unfolding()
        # Horizon adjustment
        if self.__list_variant_constraints:
            max_step = min(max_step, len(self.__list_variant_constraints)-1) ## TODO:  -1 ?

        if self.__invariant_property and not self.__final_property:
            # 1 satisfiability test is made only after all shift() calls for each step
            # Shift all remaining steps
            while self.__current_step <= max_step:
                self.shift()
            return self.__constraints_satisfied()

        else:
            # 1 test is made for each remaining step after each shift()

            # Check and correct for incoherent data on step number
            # __steps_before_check can't be >= max_step, otherwise
            # __current_step could be > max_step due to shift() of "auto shift loop"
            # and this function will not search any solution,
            # and will return an empty list.
            if self.__steps_before_check >= max_step:
                self.__steps_before_check = max_step - 1

            # Auto shift loop: Shift without checking (optimization)
            # Do not search solutions before the specified limit
            while self.__current_step < self.__steps_before_check:
                LOGGER.info("Satis test:: OPTI AUTO SHIFT")
                self.shift()

            satisfiability_test = False

            while self.__current_step < max_step:
                # Shift the system
                self.shift()

                # Test satisfiability
                satisfiability_test = self.__constraints_satisfied()
                LOGGER.info("Satis test:: done, current_step %s", self.__current_step)

                if satisfiability_test:
                    return True

            # Return the satisfiability status in all cases
            return satisfiability_test

    def squery_solve(self, vvars, max_step, max_sol):
        """Search max_sol number of solutions with max_step steps, involving vvars
        (frontiers) variables of interest.

        Assert a query is loaded (start_condition, inv_condition, final_condition)

        If __invariant_property is set and __final_property is not set,
        1 unique solution search is made only after all shift() calls for each
        remaining step from the current step.

        Otherwise, 1 test is made for each remaining step after each shift() call.

        .. note:: This function checks and corrects incoherent data on step number.
            __steps_before_check can't be >= max_step

        :param max_step: bounded horizon for computations
        :param max_sol: The maximum number of solutions to be returned
        :param vvars: Variables for which solutions must differ.
            In practice, it is a list with the values (Dimacs code)
            of all frontier places of the system.
        :type max_step: <int>
        :type max_sol: <int>
        :type vvars: <list <int>>
        :return: RawSolution objects
            RawSolution objects contain a solution got from SAT solver with all
            variable parameters from the unfolder
        :rtype: <tuple <RawSolution>>
        """
        if LOGGER.getEffectiveLevel() == DEBUG:
            LOGGER.debug("squery_solve :: vvars : %s", vvars)

        # Initialization
        self.init_forward_unfolding()
        # Horizon adjustment
        if self.__list_variant_constraints:
            max_step = min(max_step, len(self.__list_variant_constraints)-1) ## TODO:  -1 ?

        if self.__invariant_property and not self.__final_property:
            # Shift all remaining steps
            while self.__current_step <= max_step:
                self.shift()
            # Solve the problem
            return self.__msolve_constraints(max_sol, vvars)

        else:
            # Check and correct for incoherent data on step number
            # __steps_before_check can't be >= max_step, otherwise
            # __current_step could be > max_step due to shift() of "auto shift loop"
            # and this function will not search any solution,
            # and will return an empty list.
            if self.__steps_before_check >= max_step:
                self.__steps_before_check = max_step - 1

            # Auto shift loop: Shift without checking (optimization)
            # Do not search solutions before the specified limit
            print("squery_solve: begining; current_step:", self.__current_step, "max_step:", max_step)
            while self.__current_step < self.__steps_before_check:
                LOGGER.info("Solve: OPTI AUTO SHIFT")
                self.shift()

            # Search for solutions for each remaining step
            # from __current_step = __steps_before_check
            # (__current_step is incremented by the last shift() above)...
            # ... to max_step excluded.
            # Ex: __steps_before_check = 3, __current_step = 3, max_step = 4
            # => we search solutions only for 1 step (step 4)
            raw_solutions = list()
            while self.__current_step < max_step:
                self.shift()

                temp_raw_solutions = self.__msolve_constraints(max_sol, vvars)
                print("Nb raw_solutions:", len(temp_raw_solutions),
                    "steps_before_check", self.__steps_before_check,
                    "max_step", max_step,
                    "current_step", self.__current_step
                )
                #raw_input("pause")
                # When __steps_before_check is 0 by default, or less than
                # (max_step - 1), we need to search solutions for every steps
                # until max_step is reached.
                # Otherwise, __steps_before_check is set to (maxstep - 1) only
                # to allow us to search solutions for the last step (max_step).

                if temp_raw_solutions:
                    raw_solutions += temp_raw_solutions

            return raw_solutions


###############################################################################
class CLPropertyVisitor(object):
    """
    SigExpression on states and events into a set of clauses
    Type checking must be done.
    """

    def __init__(self, aux_cpt=0):
        self.cpt = aux_cpt         # for auxiliary variable naming
        self.clauses = []          # generated clauses
        self.top = True            # root of the formula?

    def visit_sig_ident(self, tnode):
        """
        ident -> literal
        """
        name = tnode.name
        new_lit = Literal(name, True)
        if self.top:
            self.clauses.append(Clause([new_lit]))
            return None
        return new_lit

    def visit_sig_not(self, node):
        """
        translate a not
        """
        top = self.top
        self.top = False
        # compile operand to nl <=> formula
        newl = node.operand.accept(self)
        notnl = newl.lit_not()
        if top: # we are at the root
            self.clauses.append(Clause([notnl]))
            return None
        return notnl # if we are not at the root

    def visit_sig_sync(self, binnode):
        """
        translate a binary (or/and) expression
        """
        top = self.top
        self.top = False
        # the internal nodes will be affected with the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)
        notnl = Literal(name, False)
        # recursive visits
        operator = binnode.operator
        nl1 = binnode.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = binnode.right_h.accept(self)
        notnl2 = nl2.lit_not()
        # clauses generation
        if operator == 'and':   # x = lh and rh
            self.clauses.extend([
                Clause([nl1, notnl]),           # not x or not lh
                Clause([nl2, notnl]),           # not x or not rh
                Clause([notnl1, notnl2, newl])  # x or not lh or not rh
            ])

        if operator == 'or':    # x = lh or rh
            self.clauses.extend([
                Clause([notnl1, newl]),         # x or not lh
                Clause([notnl2, newl]),         # x or not rh
                Clause([nl1, nl2, notnl])       # not x or not lh or not rh
            ])

        if top:
            self.clauses.append(Clause([newl]))
            return None
        return newl

    def visit_sig_const(self, exp):
        """
        translate a constant expression
        """
        top = self.top
        if top:
            if exp.is_const_false():
                raise TypeError("No constant False property allowed")
            else: # always present or satisfied
                return None
        # the internal nodes will be affected with the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)

        # clause generation
        if exp.value:
            lit = newl  # x = True
        else:
            lit = newl.lit_not()
        self.clauses.append(Clause([lit]))
        return newl

    # time operators
    def visit_sig_default(self, exp):
        """
        default translation
        """
        top = self.top
        self.top = False
        # the internal nodes will be affected the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)
        notnl = Literal(name, False)
        # recursive visits
        nl1 = exp.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = exp.right_h.accept(self)
        notnl2 = nl2.lit_not()
        # clause generation: x = lh or rh
        cl1 = Clause([notnl1, newl])        # x or not lh
        cl2 = Clause([notnl2, newl])        # x or not rh
        cl3 = Clause([nl1, nl2, notnl])     # not x or not lh or not rh

        self.clauses.append(cl1)
        self.clauses.append(cl2)
        self.clauses.append(cl3)
        if top:
            self.clauses.append(Clause([newl]))
            return None
        return newl

    def visit_sig_when(self, exp):
        """
        when translation
        """
        top = self.top
        self.top = False
        # the internal nodes will be affected the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)
        notnl = Literal(name, False)
        # recursive visits
        nl1 = exp.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = exp.right_h.accept(self)
        notnl2 = nl2.lit_not()
        # clause generation: x = lh and rh
        cl1 = Clause([nl1, notnl])           # not x or not lh
        cl2 = Clause([nl2, notnl])           # not x or not rh
        cl3 = Clause([notnl1, notnl2, newl]) # x or not lh or not rh

        self.clauses.append(cl1)
        self.clauses.append(cl2)
        self.clauses.append(cl3)
        if top:
            self.clauses.append(Clause([newl]))
            return None
        return newl




# -*- coding: utf-8 -*-
# Copyright (C) 2010-2017  IRISA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# The original code contained here was initially developed by:
#
#     Geoffroy Andrieux.
#     IRISA/IRSET
#     Symbiose team
#     IRISA  Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
#
#
# Contributor(s): Michel Le Borgne, Nolwenn Le Meur, Pierre Vignet

# Standard imports
from logging.handlers import RotatingFileHandler
import logging
import datetime as dt
import tempfile

# Custom imports
from cadbiom import info

# Paths
DIR_LOGS = tempfile.gettempdir() + '/'

# Logging
LOGGER_NAME             = info.PACKAGE_NAME
LOG_LEVEL               = 'INFO'
LOG_LEVELS              = {'debug': logging.DEBUG,
                           'info': logging.INFO,
                           'error': logging.ERROR}

################################################################################

def logger(name=LOGGER_NAME, logfilename=None):
    """Return logger of given name, without initialize it.

    Equivalent of logging.getLogger() call.

    :Example:

        cm.logger(name=__name__)

    :param name: Name of the logger, only displayed if "%(name)s" is present in
        the logging.Formatter.
    """
    return logging.getLogger(name)

_logger = logging.getLogger(LOGGER_NAME)
_logger.setLevel(LOG_LEVEL)

# log file
formatter    = logging.Formatter(
    '%(asctime)s :: %(levelname)s :: %(message)s'
)
file_handler = RotatingFileHandler(
    DIR_LOGS + LOGGER_NAME + '_' + \
    dt.datetime.now().strftime("%Y-%m-%d_%H:%M:%S") + '.log',
    'a', 100000000, 1
)
file_handler.setLevel(LOG_LEVEL)
file_handler.setFormatter(formatter)
_logger.addHandler(file_handler)

# terminal log
stream_handler = logging.StreamHandler()
formatter      = logging.Formatter('%(levelname)s: %(message)s')
stream_handler.setFormatter(formatter)
stream_handler.setLevel(LOG_LEVEL)
_logger.addHandler(stream_handler)

def log_level(level):
    """Set terminal/file log level to given one.
    .. note:: Don't forget the propagation system of messages:
        From logger to handlers. Handlers receive log messages only if
        the main logger doesn't filter them.
    """
    # Main logger
    _logger.setLevel(level.upper())
    # Handlers
    [handler.setLevel(level.upper()) for handler in _logger.handlers
        if handler.__class__ in (logging.StreamHandler,
                                 logging.handlers.RotatingFileHandler)
    ]



# -*- coding: utf-8 -*-
"""Unit tests for C API"""

from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import pytest

def test_shift_clause():

    from _cadbiom import shift_clause

    found = shift_clause([42198], 47633)
    assert found == [89831]

    found = shift_clause([66035, 66037, -66036], 47633)
    assert found == [113668, 113670, -113669]

    found = shift_clause([66036, -66037], 47633)
    assert found == [113669, -113670]

    found = shift_clause([55515, 58893, 48355, -48352], 47633)
    assert found == [103148, 106526, 95988, -95985]


def test_shift_dimacs_clauses():

    from _cadbiom import shift_dimacs_clauses

    found = shift_dimacs_clauses([[66035, 66037, -66036],
                           [66036, -66037],
                           [55515, 58893, 48355, -48352]], 47633)
    assert found == [[113668, 113670, -113669],
                     [113669, -113670],
                     [103148, 106526, 95988, -95985]]

################################################################################

def test_get_unshift_code():

    from _cadbiom import get_unshift_code

    found = get_unshift_code(-1, 47633)
    assert found == -1

    found = get_unshift_code(50, 47633)
    assert found == 50

    found = get_unshift_code(50, 2)
    assert found == 2


def test_unflatten():

    from _cadbiom import unflatten

    found = unflatten(
                [66035, 66037, -66036, 55515, 58893, 48355, -48352],
                2, # CLUnfolder.shift_step_init
                1, # CLUnfolder.shift_step_init
            )
    assert found == [[1, 1], [1, -1], [-1, 1], [1, 1], [1, 1], [1, -1]]

    found = unflatten(
                [66035, 66037, -66036, 55515, 58893, 48355, -48352],
                1, # CLUnfolder.shift_step_init
                1, # CLUnfolder.shift_step_init
            )
    assert found == [[1], [1], [-1], [1], [1], [1]]

################################################################################

@pytest.fixture()
def feed_forward_code():

    class Clause():
        def __init__(self, lits):
            self.literals = lits

    class Literal():
        def __init__(self, name, sign):
            self.name = name # string
            self.sign = sign # bool
        def __repr__(self):
            return "name:{}; sign:{}".format(self.name, self.sign)

    # Make a Clause with Literals objects
    literal_defs = [('Hsp90', True), ('not _lit2', False), ('Hsp90`', False)]
    literals = [Literal(name, sign) for name, sign in literal_defs]
    #print(literals)
    clause = Clause(literals)

    var_code_table = \
        {'Hsp90': 10,
         'not _lit2': 11,
         #'Hsp90`': 12, # volontary removed
         'NRAMP1_gene': 1,
         'PDK1_active_p': 42583,
         'EGFR_EGFR_EGF_EGF_GRB2_PAK1_active_intToMb': 2,
         'CCL26_exCellRegion_gene': 3,
         '_lit3093': 34992,
         '_lit25839': 4,
         '_lit25838': 5,
         '_lit25837': 6,
         '_lit25836': 7,
         '_lit25835': 8,
         }

    # Optimization: for forward_init_dynamic() and forward_code() (mostly
    # implemented in C.
    # Precompute __var_code_table with the literal names in "future" format;
    # i.e with a last char '`' at the end.
    # Thus it is easy to search them directly in the dict (in O(1)),
    # instead of slicing them.
    # See CLUnfolder constructor.
    temp_var_code_table = \
        {var_name + "`": var_num for var_name, var_num in var_code_table.iteritems()}
    var_code_table.update(temp_var_code_table)

    shift_step = 1

    return clause, var_code_table, shift_step


def test_forward_code(feed_forward_code):

    from _cadbiom import forward_code

    found = forward_code(feed_forward_code[0], # clause
                         feed_forward_code[1], # var_code_table
                         feed_forward_code[2]) # shift_step
    assert found == [10, -11, -11]


def test_forward_init_dynamic(feed_forward_code):

    from _cadbiom import forward_init_dynamic

    found = forward_init_dynamic([feed_forward_code[0], feed_forward_code[0]], # clause
                                 feed_forward_code[1], # var_code_table
                                 feed_forward_code[2]) # shift_step
    assert found == [[[10, -11, -11], [10, -11, -11]]]

